import unittest

from rbceq2.core_logic.alleles import Allele, BloodGroup
from rbceq2.core_logic.utils import (
    apply_to_dict_values,
    check_available_variants,
    chunk_geno_list_by_rank,
    get_non_refs,
    sub_alleles_relationships,
)



class TestCheckAvailableVariants(unittest.TestCase):
    def test_all_variants_available(self):
        """Test all variants available"""
        allele = Allele("A1", "M", ".", ".", frozenset({"A", "B", "C"}), False, 1)
        variant_pool = {"A": 2, "B": 2, "C": 2}
        result = check_available_variants(1, variant_pool, lambda x, y: x >= y, allele)
        self.assertTrue(all(result))

    def test_some_variants_unavailable(self):
        """Test some variants unavailable"""
        allele = Allele("A2", "N", ".", ".", frozenset({"A", "B", "C"}),False, 1)
        variant_pool = {"A": 0, "B": 2, "C": 1}
        result = check_available_variants(1, variant_pool, lambda x, y: x >= y, allele)
        self.assertFalse(all(result))

    def test_no_variants_available(self):
        allele = Allele("A3", "O", ".", ".", frozenset({"A", "B", "C"}),False, 1)
        variant_pool = {"A": 0, "B": 0, "C": 0}
        result = check_available_variants(1, variant_pool, lambda x, y: x >= y, allele)
        self.assertFalse(all(result))

    def test_variants_exceeding_count(self):
        allele = Allele("A4", "P", ".", ".", frozenset({"A", "B"}),False, 1)
        variant_pool = {"A": 3, "B": 5}
        result = check_available_variants(2, variant_pool, lambda x, y: x > y, allele)
        self.assertTrue(all(result))


class TestApplyToDictValues(unittest.TestCase):
    def test_modify_blood_groups(self):
        @apply_to_dict_values
        def update_genotypes(
            blood_groups: list[BloodGroup], new_genotype: str
        ) -> list[BloodGroup]:
            for bg in blood_groups:
                bg.genotypes.append(new_genotype)
            return blood_groups

        blood_group1 = BloodGroup("A", alleles={}, sample="Sample1")
        blood_group2 = BloodGroup("B", alleles={}, sample="Sample2")
        input_dict = {"group1": [blood_group1], "group2": [blood_group2]}

        expected_group1_genotypes = ["newType"]
        expected_group2_genotypes = ["newType"]

        updated_dict = update_genotypes(input_dict, "newType")
        self.assertEqual(updated_dict["group1"][0].genotypes, expected_group1_genotypes)
        self.assertEqual(updated_dict["group2"][0].genotypes, expected_group2_genotypes)

    def test_empty_dictionary(self):
        @apply_to_dict_values
        def update_phenotypes(
            blood_groups: list[BloodGroup], new_phenotype: str
        ) -> list[BloodGroup]:
            for bg in blood_groups:
                bg.phenotypes.append(new_phenotype)
            return blood_groups

        self.assertEqual(update_phenotypes({}), {})


class TestGetNonRefs(unittest.TestCase):
    def setUp(self):
        self.alleles = [
            Allele(
                "genotype1",
                "phenotype1",
                ".",
                ".",
                frozenset(["variant1"]),
                False,
                1,
                False,
                "subtype1",
            ),
            Allele(
                "genotype2",
                "phenotype2",
                ".",
                ".",
                frozenset(["variant2"]),
                False,
                1,
                True,
                "subtype2",
            ),
            Allele(
                "genotype3",
                "phenotype3",
                ".",
                ".",
                frozenset(["variant3"]),
                False,
                1,
                False,
                "subtype3",
            ),
        ]

    def test_no_reference_alleles(self):
        options = [self.alleles[0], self.alleles[2]]
        result = get_non_refs(options)
        self.assertEqual(
            result, options, "Should return all alleles when no allele is a reference"
        )

    def test_some_reference_alleles(self):
        result = get_non_refs(self.alleles)
        expected = [self.alleles[0], self.alleles[2]]
        self.assertEqual(result, expected, "Should filter out reference alleles")

    def test_all_reference_alleles(self):
        options = [
            Allele(
                genotype="genotype4",
                phenotype="phenotype4",
                genotype_alt=".",
                phenotype_alt=".",
                defining_variants=frozenset(["variant4"]),
                null=False,
                weight_geno=1,
                reference=True,
                sub_type="subtype4",
            )
        ]
        result = get_non_refs(options)
        self.assertEqual(
            result, [], "Should return an empty list when all alleles are references"
        )

    def test_empty_list(self):
        result = get_non_refs([])
        self.assertEqual(result, [], "Should handle empty input list gracefully")


class TestChunkGenoListByRank(unittest.TestCase):
    def setUp(self):
        self.alleles = [
            Allele(
                genotype="genotype1",
                phenotype="phenotype1",
                genotype_alt=".",
                phenotype_alt=".",
                defining_variants=frozenset(["variant1"]),
                null=False,
                weight_geno=1,
                reference=False,
                sub_type="A",
            ),
            Allele(
                genotype="genotype2",
                phenotype="phenotype2",
                genotype_alt=".",
                phenotype_alt=".",
                defining_variants=frozenset(["variant2"]),
                null=False,
                weight_geno=2,
                reference=False,
                sub_type="A",
            ),
            Allele(
                genotype="genotype3",
                phenotype="phenotype3",
                genotype_alt=".",
                phenotype_alt=".",
                defining_variants=frozenset(["variant3"]),
                null=False,
                weight_geno=1,
                reference=False,
                sub_type="B",
            ),
            Allele(
                genotype="genotype4",
                phenotype="phenotype4",
                genotype_alt=".",
                phenotype_alt=".",
                defining_variants=frozenset(["variant4"]),
                null=False,
                weight_geno=1,
                reference=False,
                sub_type="B",
            ),
        ]

    def test_empty_input(self):
        self.assertEqual(chunk_geno_list_by_rank([]), [])

    def test_same_weight_and_subtype(self):
        alleles = self.alleles[:2]  # Same subtype "A" with different weights
        result = chunk_geno_list_by_rank(alleles)
        self.assertEqual(len(result), 2)
        self.assertTrue(alleles[0] == result[0][0])
        self.assertTrue(alleles[1] == result[1][0])

    def test_different_weights_same_subtype(self):
        result = chunk_geno_list_by_rank(self.alleles[:2])
        self.assertEqual(len(result), 2)
        self.assertIn(self.alleles[0], result[0])
        self.assertIn(self.alleles[1], result[1])

    def test_different_subtypes_and_weights(self):
        result = chunk_geno_list_by_rank(self.alleles)
        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0]), 3)
        self.assertEqual(len(result[1]), 1)
        self.assertTrue(
            all(a in result[0] for a in [self.alleles[0]] + self.alleles[2:])
        )
        self.assertTrue(result[1][0] == self.alleles[1])

    def test_mixed_subtypes_identical_weights(self):
        alleles = [
            Allele(
                "genotype5",
                "phenotype5",
                ".",
                ".",
                frozenset(["variant5"]),
                False,
                1,
                False,
                "C",
            ),
            Allele(
                "genotype6",
                "phenotype6",
                ".",
                ".",
                frozenset(["variant6"]),
                False,
                1,
                False,
                "C",
            ),
            *self.alleles[:2],
        ]
        result = chunk_geno_list_by_rank(alleles)
        self.assertEqual(len(result), 2)
        self.assertIn(self.alleles[0], result[0])
        self.assertIn(self.alleles[1], result[1])
        self.assertIn(alleles[0], result[0])
        self.assertIn(alleles[1], result[0])


class TestSubAllelesRelationships(unittest.TestCase):
    def setUp(self):
        self.alleles = [
            Allele(
                "genotype1",
                "phenotype1",
                ".",
                ".",
                frozenset(["variant1"]),
                False,
                1,
                False,
                "subtype1",
            ),
            Allele(
                "genotype2",
                "phenotype2",
                ".",
                ".",
                frozenset(["variant1", "variant2"]),
                False,
                2,
                False,
                "subtype2",
            ),
        ]

    def test_multiple_alleles(self):
        all_alleles = {"key1": self.alleles}
        relationships, key = sub_alleles_relationships(all_alleles, "key1")
        self.assertTrue(
            relationships[f"{self.alleles[0].genotype}_isin_{self.alleles[1].genotype}"]
        )
        self.assertFalse(
            relationships[f"{self.alleles[1].genotype}_isin_{self.alleles[0].genotype}"]
        )
        self.assertEqual(key, "key1")

    def test_single_allele(self):
        all_alleles = {"key2": [self.alleles[0]]}
        relationships, key = sub_alleles_relationships(all_alleles, "key2")
        self.assertFalse(
            relationships[f"{self.alleles[0].genotype}_isin_{self.alleles[0].genotype}"]
        )
        self.assertEqual(key, "key2")

    def test_no_alleles(self):
        all_alleles = {"key3": []}
        relationships, key = sub_alleles_relationships(all_alleles, "key3")
        self.assertEqual(relationships, {})
        self.assertEqual(key, "key3")

    def test_nonexistent_key(self):
        all_alleles = {"key4": self.alleles}
        with self.assertRaises(KeyError, msg="Expected KeyError for non-existent key"):
            sub_alleles_relationships(all_alleles, "nonexistent_key")


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import argparse
import sys
from collections import defaultdict
from functools import partial
from multiprocessing import Pool
from pathlib import Path
from typing import Callable

import pandas as pd
from icecream import ic
from loguru import logger


import rbceq2.core_logic.co_existing as co
import rbceq2.core_logic.data_procesing as dp
import rbceq2.filters.geno as filt
import rbceq2.filters.phased as filt_phase
import rbceq2.filters.knops as filt_co
import rbceq2.phenotype.choose_pheno as ph
from rbceq2.core_logic.constants import PhenoType,DB_VERSION, VERSION
from rbceq2.core_logic.utils import compose, get_allele_relationships
from rbceq2.db.db import Db, prepare_db, DbDataConsistencyChecker
from rbceq2.IO.PDF_reports import generate_all_reports
from rbceq2.IO.record_data import (
    check_VCF,
    configure_logging,
    log_validation,
    record_filtered_data,
    save_df,
    stamps,
)
from rbceq2.IO.vcf import (
    VCF,
    filter_VCF_to_BG_variants,
    read_vcf,
    check_if_multi_sample_vcf,
    split_vcf_to_dfs,
)


def parse_args(args: list[str]) -> argparse.Namespace:
    """Parse command-line arguments for somatic variant calling.

    Args:
        args (List[str]): List of strings representing the command-line arguments.

    Returns:
        argparse.Namespace: An object containing the parsed command-line options.

    This function configures and interprets command-line options for a somatic
    variant caller. It expects paths to VCF files, a database file, and allows
    specification of output options and genomic references.
    """
    parser = argparse.ArgumentParser(
        description="Calls ISBT defined alleles from VCF/s",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        usage="rbceq2 --vcf example.vcf.gz --out example --reference_genome GRCh37",
    )
    version_str = f"%(prog)s {VERSION} (DB: {DB_VERSION})"


    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=version_str,
        help="Show program's version number and exit.",
    )
    parser.add_argument(
        "--vcf",
        type=lambda p: Path(p).absolute(),
        help=(
            "Path to VCF file/s. Give a folder if you want to pass multiple "
            "separate files (file names must end in .vcf or .vcf.gz), or "
            "alternatively give a file if using a multi-sample VCF."
        ),
    )
    parser.add_argument(
        "--out", type=lambda p: Path(p).absolute(), help="Prefix for output files"
    )
    parser.add_argument(
        "--depth", type=int, help="Minimum number of reads for a variant", default=10
    )
    parser.add_argument(
        "--quality",
        type=int,
        help="Minimum average genotype quality for a variant",
        default=10,
    )
    parser.add_argument(
        "--processes",
        type=int,
        help=(
            "Number of processes. I.e., how many CPUs are available? ~1GB RAM required "
            "per process"
        ),
        default=1,
    )
    parser.add_argument(
        "--reference_genome",
        type=str,
        help=("GRCh37/8"),
        choices=["GRCh37", "GRCh38"],
        required=True,
    )
    parser.add_argument(
        "--phased",
        action="store_true",
        help="Use phase information",
        default=False,
    )
    parser.add_argument(
        "--microarray",
        action="store_true",
        help="Input is from a microarray.",
        default=False,
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Enable debug logging. If not set, logging will be at info level.",
        default=False,
    )
    parser.add_argument(
        "--validate",
        action="store_true",
        help="Enable VCF validation. Doubles run time. Might help you identify input issues",
        default=False,
    )
    parser.add_argument(
        "--PDFs",
        action="store_true",
        help="Generate a per sample PDF report",
        default=False,
    )
    parser.add_argument(
        "--HPAs",
        action="store_true",
        help="Generate results for HPA",
        default=False,
    )
    # parser.add_argument(
    #     "--RH",
    #     action="store_true",
    #     help="Generate results for RHD and RHCE. WARNING! Based on SNV and small indel only - completely wrong sometimes!",
    #     default=False,
    # )

    return parser.parse_args(args)


def main():
    ic("Running RBCeq2...")

    start = pd.Timestamp.now()
    args = parse_args(sys.argv[1:])
    exclude = ["C4A", "C4B", "ATP11C", "CD99", "RHD", "RHCE"]
    # if not args.RH:
    #     exclude += ["RHD", "RHCE"]
    if not args.HPAs:
        exclude += [f"HPA{i}" for i in range(50)]
    # Configure logging
    UUID = configure_logging(args)

    logger.debug("Logger configured for debug mode.")
    logger.info("Application started.")

    # 1. Prepare the DataFrame
    logger.info("Preparing database DataFrame...")
    db_df = prepare_db()
    logger.info("Database DataFrame prepared.")

    # 2. Run consistency checks on the prepared DataFrame
    DbDataConsistencyChecker.run_all_checks(
        df=db_df, ref_genome_name=args.reference_genome
    )
    # If any check fails, an exception will be raised here, and the program will halt.

    # 3. If all checks pass, proceed to create the Db object
    logger.info("Consistency checks passed. Initializing Db object...")
    db = Db(ref=args.reference_genome, df=db_df)
    logger.info("Db object initialized.")

    if args.vcf.is_dir():
        patterns = ["*.vcf", "*.vcf.gz"]
        vcfs = [file for pattern in patterns for file in args.vcf.glob(pattern)]
        logger.info(f"{len(vcfs)} single sample VCF/s passed")
        if args.validate:
            with Pool(processes=int(args.processes)) as pool:
                for result_valid, file_name in pool.imap_unordered(
                    check_VCF, list(vcfs)
                ):
                    log_validation(result_valid, file_name)
    else:
        if args.validate:
            result_valid, file_name = check_VCF(args.vcf)
            log_validation(result_valid, file_name)
        actually_multi_vcf = check_if_multi_sample_vcf(args.vcf)
        if actually_multi_vcf:
            multi_vcf = read_vcf(args.vcf)
            logger.info("Multi sample VCF passed")
            filtered_multi_vcf = filter_VCF_to_BG_variants(
                multi_vcf, db.unique_variants
            )
            vcfs = split_vcf_to_dfs(filtered_multi_vcf)
            time_str = stamps(start)
            logger.info(f"VCFs loaded in {time_str}")
            print(f"VCFs loaded in {time_str}")
        else:
            logger.info("1 single sample VCF passed")
            vcfs = [args.vcf]

    all_alleles = defaultdict(list)
    for a in db.make_alleles():
        all_alleles[a.blood_group].append(a)
    allele_relationships = get_allele_relationships(all_alleles, int(args.processes))
    dfs_geno = {}
    dfs_pheno_numeric = {}
    dfs_pheno_alphanumeric = {}
    with Pool(processes=int(args.processes)) as pool:
        find_hits_db = partial(
            find_hits,
            db,
            args=args,
            allele_relationships=allele_relationships,
            excluded=exclude,
        )
        for results in pool.imap_unordered(find_hits_db, list(vcfs)):
            if results is not None:
                sample, genos, numeric_phenos, alphanumeric_phenos, res = results
                dfs_geno[sample] = genos
                dfs_pheno_numeric[sample] = numeric_phenos
                dfs_pheno_alphanumeric[sample] = alphanumeric_phenos
                record_filtered_data(results)
                sep = "##############"
                logger.debug(f"\n {sep} End log for sample: {sample} {sep}\n")

    df_geno = pd.DataFrame.from_dict(dfs_geno, orient="index")
    save_df(df_geno, f"{args.out}_geno.tsv", UUID)
    df_pheno_numeric = pd.DataFrame.from_dict(dfs_pheno_numeric, orient="index")
    save_df(df_pheno_numeric, f"{args.out}_pheno_numeric.tsv", UUID)
    df_pheno_alpha = pd.DataFrame.from_dict(dfs_pheno_alphanumeric, orient="index")
    save_df(df_pheno_alpha, f"{args.out}_pheno_alphanumeric.tsv", UUID)
    if args.PDFs:
        generate_all_reports(df_geno, df_pheno_alpha, df_pheno_numeric, args.out, UUID)

    time_str = stamps(start)
    logger.info(f"{len(dfs_geno)} VCFs processed in {time_str}")
    print(f"{len(dfs_geno)} VCFs processed in {time_str}")


def find_hits(
    db: Db,
    vcf: tuple[pd.DataFrame, str] | Path,
    args: argparse.Namespace,
    allele_relationships: dict[str, dict[str, bool]],
    excluded: list[str],
) -> pd.DataFrame | None:
    vcf = VCF(vcf, db.lane_variants, db.unique_variants)

    res = dp.raw_results(db, vcf, excluded)
    res = dp.make_blood_groups(res, vcf.sample)

    pipe: list[Callable] = [
        partial(
            dp.remove_alleles_with_low_read_depth,
            variant_metrics=vcf.variants,
            min_read_depth=args.depth,
            microarray=args.microarray,
        ),
        partial(
            dp.remove_alleles_with_low_base_quality,
            variant_metrics=vcf.variants,
            min_base_quality=args.quality,
            microarray=args.microarray,
        ),
        partial(dp.make_variant_pool, vcf=vcf),
        partial(
            dp.add_phasing,
            phased=args.phased,
            variant_metrics=vcf.variants,
            phase_sets=vcf.phase_sets,
        ),
        partial(filt_phase.remove_unphased, phased=args.phased),
        partial(dp.process_genetic_data, reference_alleles=db.reference_alleles),
        partial(
            dp.find_what_was_excluded_due_to_rank,
            reference_alleles=db.reference_alleles,
        ),
        filt.cant_not_include_null,
        partial(
            filt.filter_pairs_on_antithetical_zygosity, antitheticals=db.antitheticals
        ),
        partial(
            filt_phase.filter_pairs_by_phase,
            phased=args.phased,
            reference_alleles=db.reference_alleles,
        ),
        co.homs,
        co.max_rank,
        partial(co.prep_co_putative_combos, allele_relationships=allele_relationships),
        co.add_co_existing_alleles,
        partial(
            co.add_co_existing_allele_and_ref, reference_alleles=db.reference_alleles
        ),
        co.filter_redundant_pairs,
        co.mush,
        partial(
            co.list_excluded_co_existing_pairs, reference_alleles=db.reference_alleles
        ),
        partial(
            filt_co.filter_coexisting_pairs_on_antithetical_zygosity,
            antitheticals=db.antitheticals,
        ),
        partial(filt_co.remove_unphased_co, phased=args.phased),
        filt.cant_pair_with_ref_cuz_trumped,
        partial(
            filt.antithetical_modifying_SNP_is_HOM,
            antitheticals=db.antitheticals,
        ),
        filt.ensure_HET_SNP_used,
        filt.ABO_cant_pair_with_ref_cuz_261delG_HET,
        filt.cant_pair_with_ref_cuz_SNPs_must_be_on_other_side,
        filt.filter_HET_pairs_by_weight,
        filt.filter_pairs_by_context,
        filt.impossible_alleles,
        partial(filt_phase.impossible_alleles_phased, phased=args.phased),
        partial(filt_phase.filter_if_all_HET_vars_on_same_side_and_phased, phased=args.phased),
        partial(filt_phase.filter_on_in_relationship_if_HET_vars_on_dif_side_and_phased, phased=args.phased),
        filt_co.ensure_co_existing_HET_SNP_used,
        filt_co.filter_co_existing_pairs,
        filt_co.filter_co_existing_in_other_allele,
        filt_co.filter_co_existing_with_normal,  # has to be after normal filters!!!!!!!
        filt_co.filter_co_existing_subsets,
        #partial(filt_co.filter_impossible_coexisting_alleles_phased, phased=args.phased),
        dp.get_genotypes,
        dp.add_CD_to_XG,
    ]
    preprocessor = compose(*pipe)
    res = preprocessor(res)

    res = dp.add_refs(db, res, excluded)

    #merge FUT 1 and 2
    fut2s = res["FUT2"].genotypes.copy()
    fut1s = res["FUT1"].genotypes.copy()
    for allele_pair in fut2s:
        res["FUT1"].genotypes.append(allele_pair)
    for allele_pair in fut1s:
        res["FUT2"].genotypes.append(allele_pair)

    formated_called_genos = {k: ",".join(bg.genotypes) for k, bg in res.items()}

    pipe2: list[Callable] = [
        partial(ph.add_ref_phenos, df=db.df),
        partial(ph.instantiate_antigens, ant_type=PhenoType.numeric),
        partial(ph.instantiate_antigens, ant_type=PhenoType.alphanumeric),
        partial(ph.get_phenotypes1, ant_type=PhenoType.numeric),
        partial(ph.get_phenotypes1, ant_type=PhenoType.alphanumeric),
        partial(ph.get_phenotypes2, ant_type=PhenoType.numeric),
        partial(ph.get_phenotypes2, ant_type=PhenoType.alphanumeric),
        partial(ph.internal_anithetical_consistency_HET, ant_type=PhenoType.numeric),
        partial(
            ph.internal_anithetical_consistency_HET, ant_type=PhenoType.alphanumeric
        ),
        partial(ph.internal_anithetical_consistency_HOM, ant_type=PhenoType.numeric),
        partial(
            ph.internal_anithetical_consistency_HOM, ant_type=PhenoType.alphanumeric
        ),
        partial(ph.include_first_antithetical_pair, ant_type=PhenoType.numeric),
        partial(ph.include_first_antithetical_pair, ant_type=PhenoType.alphanumeric),
        partial(ph.sort_antigens, ant_type=PhenoType.numeric),
        partial(ph.sort_antigens, ant_type=PhenoType.alphanumeric),
        partial(ph.phenos_to_str, ant_type=PhenoType.numeric),
        partial(ph.phenos_to_str, ant_type=PhenoType.alphanumeric),
        partial(ph.modify_FY, ant_type=PhenoType.numeric),
        ph.combine_anitheticals,
        partial(ph.modify_FY, ant_type=PhenoType.alphanumeric),
        partial(ph.modify_KEL, ant_type=PhenoType.alphanumeric),
        partial(ph.modify_CROM, ant_type=PhenoType.alphanumeric),
        partial(ph.re_order_KEL, ant_type=PhenoType.alphanumeric),
        partial(ph.modify_MNS, ant_type=PhenoType.alphanumeric),
        partial(ph.modify_FY2, ant_type=PhenoType.alphanumeric),
    ]

    preprocessor2 = compose(*pipe2)
    res = preprocessor2(res)
    res = ph.FUT3(res)
    res = ph.FUT1(res)

    formated_called_numeric_phenos = {
        k: " | ".join(sorted(set(bg.phenotypes[PhenoType.numeric].values())))
        for k, bg in res.items()
    }
    formated_called_alphanumeric_phenos = {
        k: " | ".join(sorted(set(bg.phenotypes[PhenoType.alphanumeric].values())))
        for k, bg in res.items()
    }

    return (
        vcf.sample,
        formated_called_genos,
        formated_called_numeric_phenos,
        formated_called_alphanumeric_phenos,
        res,
    )


if __name__ == "__main__":
    main()



from __future__ import annotations

from rbceq2.core_logic.alleles import BloodGroup, Pair
from rbceq2.core_logic.constants import AlleleState
from rbceq2.core_logic.utils import Zygosity, apply_to_dict_values

from rbceq2.filters.shared_filter_functionality import (
    flatten_alleles,
    identify_unphased,
    check_var
)

from icecream import ic


@apply_to_dict_values
def filter_co_existing_with_normal(bg: BloodGroup) -> BloodGroup:
    """Filter co-existing allele pairs based on normal allele pairs.

    This function checks the co-existing allele pairs (AlleleState.CO) and removes
    those that are not present in the filtered normal allele pairs. Basically, if it
    was filtered out in any context it should be filtred out in all contexts.
    Allele pairs with any allele having 'mushed' as genotype_alt are ignored.

    Args:
        bg (BloodGroup): A BloodGroup object containing alleles, including normal and
            co-existing pairs, as well as filtered-out pairs.

    Returns:
        BloodGroup: The updated BloodGroup after filtering co-existing pairs.
    
    Example:

    KN*01.07/KN*01.07 was removed by ensure_HET_SNP_used, co that is carried over to the 
    co_existing set
    #Results:
    Genotypes count: 1
    Genotypes: KN*01.06/KN*01.07
    Phenotypes (numeric): 
    KN:1,-2,3,-4,5,6,7,8,-9,10,11,-12,13
    Phenotypes (alphanumeric):
    Kn(a+b-),McC(a+b+),Sl1-,Yk(a+),Vil+,Sl3+,KCAM-,KDAS+,DACY+,YCAD-,KNMB+

    #Data:
    Vars: 
    1:207782916_A_T : Homozygous
    1:207782769_ref : Homozygous
    1:207782856_A_G : Heterozygous
    1:207782931_A_G : Homozygous
    1:207782889_A_G : Homozygous

    Raw: 
    Allele 
    genotype: KN*01 
    defining_variants: 
            1:207782916_A_T
            1:207782769_ref 
    weight_geno: 1000 
    phenotype: KN:1,-2 or Kn(a+),Kn(b-) 
    reference: True 

    Allele 
    genotype: KN*01.06 
    defining_variants: 
            1:207782889_A_G
            1:207782856_A_G
            1:207782931_A_G 
    weight_geno: 1000 
    phenotype: KN:1,-2,-3,-4,6,7,-9,10 
    reference: False 

    Allele 
    genotype: KN*01.07 
    defining_variants: 
            1:207782889_A_G
            1:207782931_A_G 
    weight_geno: 1000 
    phenotype: KN:1,-2,-4,7,-9,10
    reference: False 

    Allele 
    genotype: KN*01.10 
    defining_variants: 
            1:207782931_A_G 
    weight_geno: 1000 
    phenotype: KN:1,-2,-9,10
    reference: False 

    #Filters applied:

    ensure_HET_SNP_used: 
    Pair(Genotype: KN*01.07/KN*01.07 
    Pair(Genotype: KN*01.07/KN*01.10

    filter_co_existing_with_normal: 
    Pair(Genotype: KN*01.07/KN*01.07
    """
    if bg.alleles[AlleleState.CO] is None:
        return bg
    to_remove = []
    filtered_pairs = [
        pair
        for pair in bg.alleles[AlleleState.NORMAL]
        if pair not in bg.filtered_out[AlleleState.NORMAL]
    ]
    for pair in bg.alleles[AlleleState.CO]:
        if any(a.genotype_alt == "mushed" for a in pair.alleles):
            continue
        if pair in filtered_pairs:
            continue
        to_remove.append(pair)

    if to_remove:
        bg.remove_pairs(to_remove, "filter_co_existing_with_normal", AlleleState.CO)

    return bg


def parse_bio_info2(pairs: list[Pair]) -> list[list[frozenset[str]]]:
    """Parse biological information string into a set of frozensets representing each
    side of '/'.

    Args:
        info (str): Biological information string.

    Returns:
        A set of frozensets, each representing unique substrings
    """

    return [pair.comparable for pair in pairs]

@apply_to_dict_values
def filter_co_existing_subsets(bg: BloodGroup) -> BloodGroup:
    """Filter co-existing allele pairs that are subsets of larger allele combinations.

    This filtering ensures that no combination smaller than either 2/2 or 1/3 (in this
    case) exists. For example, KN*01.-13+KN*01.06/KN*01.07 is a subset of
    KN*01.-13+KN*01.06+KN*01.12/KN*01.07 and therefore cannot exist.

    Args:
        bg (BloodGroup): A BloodGroup object containing allele pairs in various states.

    Returns:
        BloodGroup: The updated BloodGroup after filtering co-existing allele pairs.
        
    Example:
    ----------
    KN*01.06/KN*01.07+KN*01.10
    KN*01.07+KN*01.10/KN*01.07+KN*01.10 - not possible,
    KN*01.07+KN*01.10 + 207782856_A_G = KN*01.06

    KN*01.06 = 207782856_A_G, HET
              207782916_A_T,
              207782889_A_G,
              207782931_A_G
    KN*01.07 = 207782889_A_G,
              207782916_A_T
    KN*01.10 = 207782931_A_G,
              207782916_A_T

    variant_pool = {
        '1:207782856_A_G': 'Heterozygous',
        '1:207782889_A_G': 'Homozygous',
        '1:207782916_A_T': 'Homozygous',
        '1:207782931_A_G': 'Homozygous'
    }
    """

    def compare_to_all(pair_to_check: Pair, all_co_pairs: list[Pair]) -> bool:
        """
        Compare a pair against a list of other pairs to check if it is a proper subset.
        """
        # Flatten the allele genotypes from the pair we are checking into a single set.
        # e.g., Pair('A', 'B+C') -> {'A', 'B', 'C'}
        flat_alleles_to_check = {g for side in pair_to_check.comparable for g in side}

        for other_pair in all_co_pairs:
            if pair_to_check == other_pair:
                continue

            flat_other_alleles = {g for side in other_pair.comparable for g in side}

            # --- THE CORE FIX IS HERE ---
            # Use the proper subset operator '<' on the flattened sets of allele names.
            # This ensures the pair is strictly smaller than the one it's compared to,
            # which resolves the symmetrical removal bug.
            if flat_alleles_to_check < flat_other_alleles:
                # This secondary logic for HET variants is preserved from the original.
                different_alleles = [
                    allele
                    for allele in bg.alleles[AlleleState.RAW]
                    if allele.genotype in flat_other_alleles.difference(flat_alleles_to_check)
                ]
                alleles_with_1_HET_var = [
                    allele
                    for allele in different_alleles
                    if [
                        bg.variant_pool.get(variant) for variant in allele.defining_variants
                    ].count(Zygosity.HET)
                    == 1
                ]
                if alleles_with_1_HET_var and not pair_to_check.same_subtype:
                    continue
                
                # If the checks pass, we've found a superset, so this pair should be removed.
                return True
        
        return False

    if bg.alleles[AlleleState.CO] is None:
        return bg

    to_remove = []
    all_co_pairs = bg.alleles[AlleleState.CO]
    all_co_pairs_without_ref = [pair for pair in all_co_pairs if not pair.contains_reference]
    
    for pair in all_co_pairs:
        if pair.contains_reference:
            if compare_to_all(pair, all_co_pairs):
                to_remove.append(pair)
        elif compare_to_all(pair, all_co_pairs_without_ref):
            to_remove.append(pair)

    if to_remove:
        bg.remove_pairs(to_remove, "filter_co_existing_subsets", AlleleState.CO)

    return bg

#this was replaced by the above (AI generated) version after unit tests detected a logic flaw:
# You are absolutely correct. Thank you for providing the class definitions. This context is critical, and you've pinpointed the exact area of complexity that explains the behavior. My previous analysis was incomplete because it was missing this context.
# The issue is a fascinating and subtle conflict between two different ways of defining a "subset" relationship:
# The Scientific Definition (Allele.__contains__): The Allele class defines the in operator based on defining_variants. allele_A in allele_B is True only if allele_A's variants are a proper subset of allele_B's variants. This is the scientifically correct and intuitive way to determine if one allele is a simpler version of another.
# The String-Based Logic (in the filter): The filter_co_existing_subsets function, as written, completely ignores the Allele.__contains__ method. Instead, it uses its own internal helper function, is_subset, which operates on the Pair.comparable property. This property is derived from the allele's genotype string, splitting it by +.
# This is the root of the problem. The test fails because the function's internal logic for determining a "subset" is flawed and does not behave like a true mathematical subset operation.
# The Flaw in the Code's Logic
# Let's re-examine the test case with this new understanding:
# pair_subset: Contains alleles A and B. Its comparable property effectively represents the sets {'A'} and {'B'}.
# pair_superset: Contains alleles A+B and B. Its comparable property represents the sets {'A', 'B'} and {'B'}.
# When the function's internal is_subset logic runs:
# It compares pair_subset to pair_superset: It checks if the elements {'A'} and {'B'} from the subset exist within the flattened elements of the superset ({'A', 'B'}). They do. So, pair_subset is correctly marked for removal.
# It compares pair_superset to pair_subset: It checks if the elements {'A', 'B'} and {'B'} from the superset exist within the flattened elements of the subset ({'A', 'B'}). They do. The flawed logic concludes that the superset is a "subset" of the subset and incorrectly marks pair_superset for removal as well.
# The function should only remove a pair if it is a proper subset of another. The current implementation removes any pair that has overlapping allele components, leading to the incorrect symmetrical removal that your test caught.
# The Suggested Code Change
# The best way to fix the function is to make its internal logic behave like a proper subset check, respecting its design of working with the flattened genotype strings. We can achieve this by comparing the size of the flattened allele sets. A pair is only a subset if its total set of allele components is a proper subset of the other's.
# Here is the corrected version of the filter_co_existing_subsets function. This version fixes the flawed logic, and with this change, your original unit test will pass as expected.
# E2e tests seem fine with it but I'll keep this here for a few iterations just in case 

# @apply_to_dict_values  
# def filter_co_existing_subsets(bg: BloodGroup) -> BloodGroup:
#     """Filter co-existing allele pairs that are subsets of larger allele combinations.

#     This filtering ensures that no combination smaller than either 2/2 or 1/3 (in this
#     case) exists. For example, KN*01.-13+KN*01.06/KN*01.07 is a subset of
#     KN*01.-13+KN*01.06+KN*01.12/KN*01.07 and therefore cannot exist.

#     Args:
#         bg (BloodGroup): A BloodGroup object containing allele pairs in various states.

#     Returns:
#         BloodGroup: The updated BloodGroup after filtering co-existing allele pairs.
#      Example:
#     ----------
#     KN*01.06/KN*01.07+KN*01.10
#     KN*01.07+KN*01.10/KN*01.07+KN*01.10 - not possible,
#     KN*01.07+KN*01.10 + 207782856_A_G = KN*01.06

#     KN*01.06 = 207782856_A_G, HET
#               207782916_A_T,
#               207782889_A_G,
#               207782931_A_G
#     KN*01.07 = 207782889_A_G,
#               207782916_A_T
#     KN*01.10 = 207782931_A_G,
#               207782916_A_T

#     variant_pool = {
#         '1:207782856_A_G': 'Heterozygous',
#         '1:207782889_A_G': 'Homozygous',
#         '1:207782916_A_T': 'Homozygous',
#         '1:207782931_A_G': 'Homozygous'
#     }
#     """

#     def compare_to_all(pair_compare, comparables):
#         """Compare a pair against a list of pairs to check if it is a subset.

#         Args:
#             pair_compare (tuple): A pair representing the allele combination to compare.
#             comparables (list[tuple]): A list of allele pairs to compare against.

#         Returns:
#             bool: True if the pair is a subset of any other pair, False otherwise.
#         """

#         def number_of_alleles(pair_to_compare):
#             return sum(len(bit) for bit in pair_to_compare)

#         def is_subset(allele_to_compare):
#             return all(
#                 co_allele in other_pair[0] or co_allele in other_pair[1]
#                 for co_allele in allele_to_compare
#             )

#         for other_pair in comparables:
#             if pair_compare == other_pair:
#                 continue
#             no_alleles_in_mushed_pair = number_of_alleles(pair_compare)
#             no_alleles_in_other_mushed_pair = number_of_alleles(other_pair)
#             flat_alleles = flatten_alleles(pair_compare)
#             flat_other_alleles = flatten_alleles(other_pair)

#             if (
#                 no_alleles_in_mushed_pair == no_alleles_in_other_mushed_pair
#                 and flat_alleles == flat_other_alleles
#             ):
#                 continue
#                 # don't remove due to het permutations ie
#                 # KN*01.10', 'KN*01.-05+KN*01.07
#                 # KN*01.-05', 'KN*01.07+KN*01.10
#             if is_subset(pair_compare[0]) and is_subset(pair_compare[1]):
#                 different_alleles = [
#                     allele
#                     for allele in bg.alleles[AlleleState.RAW]
#                     if allele.genotype in flat_other_alleles.difference(flat_alleles)
#                 ]
#                 alleles_with_1_HET_var = [
#                     allele
#                     for allele in different_alleles
#                     if [
#                         bg.variant_pool[variant] for variant in allele.defining_variants
#                     ].count(Zygosity.HET)
#                     == 1
#                 ]
#                 if alleles_with_1_HET_var and not pair.same_subtype:
#                     continue
#                 return True
#         return False

#     if bg.alleles[AlleleState.CO] is None:
#         return bg
#     to_remove = []
#     all_comparable = parse_bio_info2(bg.alleles[AlleleState.CO])
#     all_comparable_without_ref = parse_bio_info2(
#         [pair for pair in bg.alleles[AlleleState.CO] if not pair.contains_reference]
#     )
#     for pair in bg.alleles[AlleleState.CO]:
#         if pair.contains_reference:
#             if compare_to_all(pair.comparable, all_comparable):
#                 to_remove.append(pair)
#         elif compare_to_all(pair.comparable, all_comparable_without_ref):
#             to_remove.append(pair)
#     if to_remove:
#         bg.remove_pairs(to_remove, "filter_co_existing_subsets", AlleleState.CO)

#     return bg


@apply_to_dict_values
def filter_coexisting_pairs_on_antithetical_zygosity(
    bg: BloodGroup, antitheticals: dict[str, list[str]]
) -> BloodGroup:
    """Process genetic data to identify alleles and genotypes.

    Args:
    ----------
    res (dict[str, list[Allele]]):
        A dictionary mapping genotypes to lists of Allele objects.
    variant_pool_numeric (dict[str, int]):
        A dictionary mapping variants to their counts.

    Returns:
    ----------
    list[str]:
        A list of called genotypes based on processed data.
    
    Example:
    
    Genotypes count: 4
    Genotypes: 
    KN*01.-05+KN*01.10/KN*02
    KN*01.-05/KN*02.10
    KN*01.10/KN*02  
    KN*01/KN*02.10
    # note, the last two options are the case where 1:207760773_C_T is with
    # 1:207782769_G_A, RBCeq2 considers this possible, albiet unlikely
    # use of phased data will resolve this

    Phenotypes (numeric): KN:1,2,3,4,5,-6,-7,8,9,10,11,-12,13
    Phenotypes (alphanumeric): 
    Kn(a+b+),McC(a+b-),Sl1+,Yk(a+),Vil-,Sl3+,KCAM+,KDAS+,DACY+,YCAD-,KNMB+

    #Data:
    Vars: 
    1:207782916_A_T : Homozygous
    1:207782769_ref : Heterozygous
    1:207760773_C_T : Heterozygous
    1:207782931_A_G : Heterozygous
    1:207782769_G_A : Heterozygous

    Raw: 
    Allele 
    genotype: KN*01 
    defining_variants: 
            1:207782769_ref
            1:207782916_A_T #HOM
    weight_geno: 1000 
    phenotype: KN:1,-2 or Kn(a+),Kn(b-) 
    reference: True 

    Allele 
    genotype: KN*01.-05 
    defining_variants: 
            1:207760773_C_T 
    weight_geno: 1000 
    phenotype: KN:1,-2,-5 or Kn(a+),Kn(b-),Yk(a-) 
    reference: False 

    Allele 
    genotype: KN*01.10 
    defining_variants: 
            1:207782931_A_G 
    weight_geno: 1000 
    phenotype: KN:1,-2,-9,10 or Kn(a+),Kn(b-),KCAM-,KDAS+ 
    reference: False 

    Allele 
    genotype: KN*02 
    defining_variants: 
            1:207782769_G_A 
    weight_geno: 1000 
    phenotype: KN:-1,2 or Kn(a-),Kn(b+) 
    reference: False 

    Allele 
    genotype: KN*02.10 
    defining_variants: 
            1:207782931_A_G
            1:207782769_G_A 
    weight_geno: 1000 
    phenotype: KN:-1,2,-9,10 or Kn(a-),Kn(b+),KCAM-,KDAS+ 
    reference: False 

    #Filters applied:

    filter_pairs_on_antithetical_zygosity: 
    Pair(Genotype: KN*01.-05/KN*01.10 
    Pair(Genotype: KN*01/KN*01.-05 
    Pair(Genotype: KN*01/KN*01.10

    filter_co_pairs_on_antithetical_zygosity: 
    Pair(Genotype: KN*01.-05/KN*01.10 
    Pair(Genotype: KN*01/KN*01.-05+KN*01.10 #the point of this filter
    Pair(Genotype: KN*01/KN*01.-05 
    Pair(Genotype: KN*01/KN*01.10

    """
    
    if bg.alleles[AlleleState.CO] is None:
        return bg
    if bg.type not in antitheticals:
        return bg
    
    to_remove = []
    flattened_sub_types = {
        allele.sub_type
        for allele in flatten_alleles(bg.alleles[AlleleState.NORMAL])
    }
    if len(flattened_sub_types) > 1:
        for pair in bg.alleles[AlleleState.CO]:
            flat_sub_types = {allele.sub_type for allele in pair}
            if flat_sub_types == flattened_sub_types:
                continue
            else:
                to_remove.append(pair)
    if to_remove:
        bg.remove_pairs(
        to_remove, "filter_co_pairs_on_antithetical_zygosity", AlleleState.CO
    )

    return bg



@apply_to_dict_values
def ensure_co_existing_HET_SNP_used(bg: BloodGroup) -> BloodGroup:
    """
    Ensures that heterozygous variants are utilized in allele pairs if they can form
    existing alleles.

    This function iterates over heterozygous variants in the variant pool of a
    BloodGroup object.
    For each heterozygous variant, it checks whether adding this variant to the defining
    variants of alleles in each pair results in alleles that already exist outside the
    pair. If multiple such matches are found, the pair is considered invalid and is
    removed from the allele pairs.

    Args:
        bg (BloodGroup): The BloodGroup object containing allele pairs and variant pool
        information.

    Returns:
        BloodGroup: The updated BloodGroup object with inconsistent allele pairs
        removed.

    Example:
        Suppose you have a BloodGroup object with allele pairs and a variant pool that
        includes heterozygous variants. This function will remove pairs where the
        heterozygous variants are not properly used in the allele definitions, but
        could form existing alleles when added.

    Genotypes count: 5
    Genotypes: 
    KN*01.-05+KN*01.10/KN*02
    KN*01.-05/KN*02##### not possible, 207782931_A_G must be on one side
    KN*01.-05/KN*02.10
    KN*01.10/KN*02
    KN*01/KN*02.10
    Phenotypes (numeric): KN:1,2,3,4,5,-6,-7,8,9,-10,11,-12,13
    KN:1,2,3,4,5,-6,-7,8,9,10,11,-12,13
    Phenotypes (alphanumeric): Kn(a+b+),McC(a+b-),Sl1+,Yk(a+),Vil-,Sl3+,KCAM+,KDAS+,DACY+,YCAD-,KNMB+
    Kn(a+b+),McC(a+b-),Sl1+,Yk(a+),Vil-,Sl3+,KCAM+,KDAS-,DACY+,YCAD-,KNMB+

    #Data:
    Vars: 
    1:207782916_A_T : Homozygous
    1:207782769_ref : Heterozygous
    1:207760773_C_T : Heterozygous
    1:207782931_A_G : Heterozygous
    1:207782769_G_A : Heterozygous
  
    Raw: 
    Allele 
    genotype: KN*01 
    defining_variants: 
            1:207782916_A_T
            1:207782769_ref 
    weight_geno: 1000 
    phenotype: KN:1,-2 or Kn(a+),Kn(b-) 
    reference: True 

    Allele 
    genotype: KN*01.-05 
    defining_variants: 
            1:207760773_C_T 
    weight_geno: 1000 
    phenotype: KN:1,-2,-5 or Kn(a+),Kn(b-),Yk(a-) 
    reference: False 

    Allele 
    genotype: KN*01.10 
    defining_variants: 
            1:207782931_A_G 
    weight_geno: 1000 
    phenotype: KN:1,-2,-9,10 or Kn(a+),Kn(b-),KCAM-,KDAS+ 
    reference: False 

    Allele 
    genotype: KN*02 
    defining_variants: 
            1:207782769_G_A 
    weight_geno: 1000 
    phenotype: KN:-1,2 or Kn(a-),Kn(b+) 
    reference: False 

    Allele 
    genotype: KN*02.10 
    defining_variants: 
            1:207782769_G_A
            1:207782931_A_G 
    weight_geno: 1000 
    phenotype: KN:-1,2,-9,10 or Kn(a-),Kn(b+),KCAM-,KDAS+ 
    reference: False 


    """
    if bg.type != "KN":
        return bg
    to_remove = []
    for variant, zygo in bg.variant_pool.items():
        if zygo == Zygosity.HET:
            for pair in bg.alleles[AlleleState.CO]:
                hits = check_var(bg, pair, AlleleState.CO, variant)
                if hits > 1:
                    to_remove.append(pair)

    if to_remove:
        bg.remove_pairs(to_remove, "ensure_co_existing_HET_SNP_used", AlleleState.CO)

    return bg


@apply_to_dict_values
def remove_unphased_co(bg: BloodGroup, phased: bool) -> BloodGroup:
    """Remove unphased alleles from the BloodGroup's bg.alleles[AlleleState.CO] 
    state if phased flag is set.

    This function iterates through the alleles in the FILT state and checks their
    phasing. Alleles with more than two distinct phases trigger a warning. If an
    allele has exactly two phases and no placeholder ('.') is present, it is marked
    for removal. Alleles with a single phase remain, as they are assumed to align
    with the reference.

    Args:
        bg (BloodGroup): A BloodGroup object containing allele states and phasing
            information.
        phased (bool): A flag indicating whether phasing should be enforced.

    Returns:
        BloodGroup: The updated BloodGroup with improperly phased alleles removed.
    
    Example:
            
    #Results:
    Genotypes count: 1
    Genotypes: KN*01.-05/KN*01.06
    Phenotypes (numeric): KN:1,-2,3,4,5,6,7,8,9,10,11,-12,13
    Phenotypes (alphanumeric): 
    Kn(a+b-),McC(a+b+),Sl1+,Yk(a+),Vil+,Sl3+,KCAM+,KDAS+,DACY+,YCAD-,KNMB+

    #Data:
    Vars: 
    1:207609424_ref : Homozygous
    1:207609571_A_T : Homozygous
    1:207587428_C_T : Heterozygous
    1:207609544_A_G : Heterozygous
    1:207609586_A_G : Heterozygous
    1:207609511_A_G : Heterozygous
    Vars_phase: 
    1:207609424_ref : 1/1
    1:207609571_A_T : 1/1
    1:207587428_C_T : 1|0
    1:207609544_A_G : 0|1
    1:207609586_A_G : 0|1
    1:207609511_A_G : 0|1
    Vars_phase_set: 
    1:207609424_ref : .
    1:207609571_A_T : .
    1:207587428_C_T : 202971590
    1:207609544_A_G : 202971590
    1:207609586_A_G : 202971590
    1:207609511_A_G : 202971590

    Raw: 
    Allele 
    genotype: KN*01 
    defining_variants: 
            1:207609571_A_T
            1:207609424_ref 
    weight_geno: 1000 
    phenotype: KN:1,-2 or Kn(a+),Kn(b-) 
    reference: True 

    Allele 
    genotype: KN*01.-05 
    defining_variants: 
            1:207587428_C_T 1|0
    weight_geno: 1000 
    phenotype: KN:1,-2,-5 or Kn(a+),Kn(b-),Yk(a-) 
    reference: False 

    Allele 
    genotype: KN*01.06 
    defining_variants: 
            1:207609586_A_G
            1:207609511_A_G
            1:207609544_A_G 
    weight_geno: 1000 
    phenotype: KN:1,-2,-3,-4,6,7,-9,10 or 
    Kn(a+),Kn(b-),Sl1-,McC(a-),McC(b+),Vil+,KCAM-,KDAS+ 
    reference: False 

    Allele 
    genotype: KN*01.07 
    defining_variants: 
            1:207609586_A_G
            1:207609544_A_G 
    weight_geno: 1000 
    phenotype: KN:1,-2,-4,7,-9,10 or
    Kn(a+),Kn(b-),Sl1-,Vil+,KCAM-,KDAS+ 
    reference: False 

    Allele 
    genotype: KN*01.10 
    defining_variants: 
            1:207609586_A_G 0|1
    weight_geno: 1000 
    phenotype: KN:1,-2,-9,10 or 
    Kn(a+),Kn(b-),KCAM-,KDAS+ 
    reference: False 

    #Filters applied:

    2025-07-24 11:42:24.409 | DEBUG    | 
    remove_unphased_co: 
    Pair(Genotype: KN*01/KN*01.-05+KN*01.06
    Pair(Genotype: KN*01/KN*01.-05+KN*01.07
    Pair(Genotype: KN*01/KN*01.-05+KN*01.10

    ie, KN*01.-05+KN*01.10 not possible:
        1:207609586_A_G = 0|1
        1:207587428_C_T 1|0
    """

    if not phased or bg.type != "KN":
        return bg
        
    co_alleles = bg.alleles[AlleleState.CO]
    if co_alleles is None:
        return bg
    to_remove = []
    alleles_to_remove = identify_unphased(bg, flatten_alleles(co_alleles))
    for pair in bg.alleles[AlleleState.CO]:
        if pair.allele1 in alleles_to_remove or pair.allele2 in alleles_to_remove:
            to_remove.append(pair)

    if to_remove:
        bg.remove_pairs(to_remove, "remove_unphased_co", AlleleState.CO)
    return bg


# doesn't come up as it was written to handle & in A4GALT
# keeping due to chance that future additions to KN will
# require same logic
@apply_to_dict_values
def filter_co_existing_pairs(bg: BloodGroup) -> BloodGroup:
    """Example:
    [Allele(genotype='A4GALT*01.02',
           genotype_alt='.',
           defining_variants=frozenset({'22:43089849_T_C'}),
           weight_geno=1000,
           reference=False,
           sub_type='A4GALT*01',
    Allele(genotype='A4GALT*02',
           genotype_alt='.',
           defining_variants=frozenset({'22:43113793_C_A'}),
           weight_geno=1000,
           reference=False,
           sub_type='A4GALT*02',
    Allele(genotype='A4GALT*02.02',
           genotype_alt='.',
           defining_variants=frozenset({'22:43089849_T_C',
                                        '22:43113793_C_A'}),
           weight_geno=1000,
           reference=False,
           sub_type='A4GALT*02',.

    variant_pool={'22:43089849_T_C': 'Heterozygous',
                   '22:43113793_C_A': 'Heterozygous'},

    filtered_out=defaultdict(<class 'list'>,  #modifying SNP has to be on one side
    {'filter_pairs_by_context': [[Allele(genotype='A4GALT*02',
                                         defining_variants=frozenset({'22:43113793_C_A'}),
                                  Allele(genotype='A4GALT*01',
                                         defining_variants=frozenset({'22:43113793_ref'}),

     'filter_pairs_on_antithetical_zygosity': #22:43113793 is HET so can't
     have A4GALT*01* on both sides
                                [[Allele(genotype='A4GALT*01.02',
                                        defining_variants=frozenset({'22:43089849_T_C'}),
                                Allele(genotype='A4GALT*01',
                                        defining_variants=frozenset({'22:43113793_ref'}),

    sample: 008Kenya A4GALT
    A4GALT*01.02/A4GALT*02
    A4GALT*01/A4GALT*02.02
     me:
     A4GALT*01.02/A4GALT*02
     A4GALT*01/A4GALT*02.02
     A4GALT*01/A4GALT*01.02&A4GALT*02 not possible:
        can't have A4GALT*01 and A4GALT*02 on same side,
        can't have ref/ref and
        have to have only A4GALT*01* on one side and only A4GALT*02* on the other side,
         when anithetical SNP HET.
        DIFFERENT logic to both existing filters
    """  # noqa: D401, D205
    if bg.alleles[AlleleState.CO] is not None:
        to_remove = []
        for pair in bg.alleles[AlleleState.CO]:
            for allele in pair:
                if allele.genotype_alt == "mushed" and "/" in allele.sub_type:
                    to_remove.append(pair)
                    break

        bg.remove_pairs(to_remove, "filter_co_existing_pairs", AlleleState.CO)

    return bg


# doesn't come up as its about & and FUT2*01.06
# keeping incase this logic is needed for KN one day
@apply_to_dict_values
def filter_co_existing_in_other_allele(bg: BloodGroup) -> BloodGroup:
    """Example:
    ----------
    1:

    226
    FUT2*01.03.01/FUT2*01.06
    FUT2*01/FUT2*01.03.03
    FUT2*01/FUT2*01.03.01&FUT2*01.06 - not possible
    FUT2*01.03.01 and FUT2*01.06 have 1 defining variant each,
    together they define FUT2*01.03.03
    """  # noqa: D401, D415, D400, D205
    if bg.alleles[AlleleState.CO] is not None:
        to_remove = []
        for pair in bg.alleles[AlleleState.CO]:
            for allele in pair:
                if allele.genotype_alt == "mushed":
                    for raw_allele in bg.alleles[AlleleState.FILT]:
                        if (
                            raw_allele.defining_variants == allele.defining_variants
                            and raw_allele.genotype != allele.genotype
                        ):
                            to_remove.append(pair)
        bg.remove_pairs(to_remove, "filter_co_existing_in_other_allele", AlleleState.CO)

    return bg



import unittest
from collections import defaultdict
from rbceq2.core_logic.utils import Zygosity
from rbceq2.core_logic.alleles import Allele, BloodGroup, Pair
from rbceq2.core_logic.constants import AlleleState
from rbceq2.filters.geno import (
    ABO_cant_pair_with_ref_cuz_261delG_HET,
    cant_pair_with_ref_cuz_SNPs_must_be_on_other_side,
    cant_pair_with_ref_cuz_trumped,
    filter_HET_pairs_by_weight,
    filter_pairs_by_context,
    filter_pairs_on_antithetical_zygosity,
    flatten_alleles,
    split_pair_by_ref,
    antithetical_modifying_SNP_is_HOM
)



class TestFlattenAlleles(unittest.TestCase):
    def test_flatten_alleles(self):
        allele1 = Allele(
            genotype="A1",
            phenotype="M",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var1"}),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="subtype1",
        )
        allele2 = Allele(
            genotype="A2",
            phenotype="N",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var2"}),
            null=False,
            weight_geno=2,
            reference=False,
            sub_type="subtype2",
        )
        pair1 = Pair(allele1=allele1, allele2=allele2)
        pair2 = Pair(allele1=allele2, allele2=allele1)
        # Should not add duplicates due to set behavior

        expected = {allele1, allele2}
        result = flatten_alleles([pair1, pair2])
        self.assertEqual(
            result, expected, "Should return a unique set of alleles from pairs"
        )

    def test_empty_list(self):
        expected = set()
        result = flatten_alleles([])
        self.assertEqual(
            result, expected, "Should return an empty set when input is an empty list"
        )

    def test_all_identical_pairs(self):
        allele = Allele(
            genotype="A1",
            phenotype="M",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var1"}),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="subtype1",
        )
        pair = Pair(allele1=allele, allele2=allele)
        expected = {allele}
        result = flatten_alleles([pair, pair])
        self.assertEqual(
            result,
            expected,
            "Should return a set with a single allele when all pairs are identical",
        )


class TestSplitPairByRef(unittest.TestCase):
    def test_normal_case(self):
        allele1 = Allele(
            genotype="A1",
            phenotype="M",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var1"}),
            null=False,
            weight_geno=1,
            reference=True,
            sub_type="subtype1",
        )
        allele2 = Allele(
            genotype="A2",
            phenotype="N",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var2"}),
            null=False,
            weight_geno=2,
            reference=False,
            sub_type="subtype2",
        )
        pair = Pair(allele1=allele1, allele2=allele2)
        ref, non_ref = split_pair_by_ref(pair)
        self.assertEqual(ref, allele1)
        self.assertEqual(non_ref, allele2)

    def test_both_reference(self):
        allele1 = Allele(
            genotype="A1",
            phenotype="M",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var1"}),
            null=False,
            weight_geno=1,
            reference=True,
            sub_type="subtype1",
        )
        allele2 = Allele(
            genotype="A2",
            phenotype="N",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var2"}),
            null=False,
            weight_geno=2,
            reference=True,
            sub_type="subtype2",
        )
        pair = Pair(allele1=allele1, allele2=allele2)
        with self.assertRaises(ValueError):
            split_pair_by_ref(pair)

    def test_neither_reference(self):
        allele1 = Allele(
            genotype="A1",
            phenotype="M",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var1"}),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="subtype1",
        )
        allele2 = Allele(
            genotype="A2",
            phenotype="N",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"var2"}),
            null=False,
            weight_geno=2,
            reference=False,
            sub_type="subtype2",
        )
        pair = Pair(allele1=allele1, allele2=allele2)
        with self.assertRaises(ValueError):
            split_pair_by_ref(pair)







class TestFilterPairsOnAntitheticalZygosity(unittest.TestCase):
    def setUp(self):
        allele2 = Allele(
            genotype="FY*02",
            phenotype="FY:2",
            genotype_alt="FY*B",
            phenotype_alt="Fy(b+)",
            defining_variants=frozenset({"1:159175354_G_A"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="FY*02",
        )
        allele3 = Allele(
            genotype="FY*01",
            phenotype="FY:1",
            genotype_alt="FY*A",
            phenotype_alt="Fy(a+)",
            defining_variants=frozenset({"1:159175354_ref"}),
            null=False,
            weight_geno=1000,
            reference=True,
            sub_type="FY*01",
        )
        allele4 = Allele(
            genotype="FY*01N.01",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt="Fy(a-b-)",
            defining_variants=frozenset({"1:159175354_ref", "1:159174683_T_C"}),
            null=False,
            weight_geno=7,
            reference=False,
            sub_type="FY*01",
        )
        # have to have both subtypes in pair
        self.pair1 = Pair(allele1=allele2, allele2=allele4)  # ok
        self.pair2 = Pair(allele1=allele3, allele2=allele4)  # not ok
        self.bg = BloodGroup(
            type="FY",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2]},
            sample="013Kenya",
            variant_pool={"1:159175354_G_A": Zygosity.HET},
            filtered_out=defaultdict(list),
        )

        self.antitheticals = {
            "KN": ["207782916_A_T", "207782769_G_A,207782916_A_T,207782931_A_G"],
            "LU": ["45315445_G_A", "45315445_ref"],
            "LW": ["10397987_ref", "10397987_A_G"],
            "SC": ["43296522_ref", "43296522_G_A"],
            "YT": ["100490797_ref", "100490797_G_T"],
            "FY": ["159175354_ref", "159175354_G_A"],
        }
        filter_pairs_on_antithetical_zygosity({1: self.bg}, self.antitheticals)

    def test_pairs_removed(self):
        self.assertTrue(
            self.pair2 in self.bg.filtered_out["filter_pairs_on_antithetical_zygosity"]
        )
        self.assertTrue(self.pair2 not in self.bg.alleles[AlleleState.NORMAL])

    def test_no_pairs_removed(self):
        self.assertTrue(
            self.pair1
            not in self.bg.filtered_out["filter_pairs_on_antithetical_zygosity"]
        )
        self.assertTrue(self.pair1 in self.bg.alleles[AlleleState.NORMAL])

    def test_empty_normal_list(self):
        bg = BloodGroup(
            type="FY",
            alleles={AlleleState.NORMAL: []},
            sample="013Kenya",
            variant_pool={"1:159175354_G_A": Zygosity.HET},
            filtered_out=defaultdict(list),
        )
        filtered_bg = list(
            filter_pairs_on_antithetical_zygosity({1: bg}, self.antitheticals).values()
        )[0]
        self.assertEqual(filtered_bg.alleles[AlleleState.NORMAL], [])


class TestFilterPairsOnAntitheticalModifyingSNP(unittest.TestCase):
    def setUp(self):
        allele1 = Allele(
            genotype="LU*02",
            phenotype="LU:2",
            genotype_alt="LU*B",
            phenotype_alt="Lu(a-b+)",
            defining_variants=frozenset({"19:45315445_ref"}),
            null=False,
            weight_geno=1,
            reference=True,
            sub_type="LU*02",
        )
        allele2 = Allele(
            genotype="LU*02.19",
            phenotype="LU:-18,19",
            genotype_alt=".",
            phenotype_alt="Au(a-b+)",
            defining_variants=frozenset({"19:45315445_ref", "19:45322744_A_G"}),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="LU*02",
        )
        allele3 = Allele(
            genotype="LU*01.19",
            phenotype="LU:..",
            genotype_alt=".",
            phenotype_alt="",
            defining_variants=frozenset({"19:45315445_G_A", "19:45322744_A_G"}),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="LU*01",
        )  #'LU*01.19/LU*02' not possible because modifying SNP (45322744_A_G) is hom
        self.pair1 = Pair(allele1=allele1, allele2=allele2)
        self.pair2 = Pair(allele1=allele1, allele2=allele3)

        self.antitheticals = {
            "KN": ["207782916_A_T", "207782769_G_A,207782916_A_T,207782931_A_G"],
            "LU": ["45315445_G_A", "45315445_ref"],
            "LW": ["10397987_ref", "10397987_A_G"],
            "SC": ["43296522_ref", "43296522_G_A"],
            "YT": ["100490797_ref", "100490797_G_T"],
        }

    def test_pairs_removed_due_to_homozygous_snp(self):
        bg = BloodGroup(
            type="LU",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2]},
            sample="128",
            variant_pool={
                "19:45315445_G_A": "Heterozygous",
                "19:45315445_ref": "Heterozygous",
                "19:45322744_A_G": "Homozygous",
            },
            filtered_out=defaultdict(list),
        )
        filtered_bg = list(
            antithetical_modifying_SNP_is_HOM(
                {1: bg}, self.antitheticals
            ).values()
        )[0]
        self.assertTrue(self.pair2 not in filtered_bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair2
            in filtered_bg.filtered_out["antithetical_modifying_SNP_is_HOM"]
        )

    def test_no_pairs_removed_due_to_heterozygous_snp(self):
        bg = BloodGroup(
            type="LU",
            alleles={AlleleState.NORMAL: [self.pair1]},
            sample="128",
            variant_pool={
                "19:45315445_G_A": "Heterozygous",
                "19:45315445_ref": "Heterozygous",
                "19:45322744_A_G": "Homozygous",
            },
            filtered_out=defaultdict(list),
        )
        filtered_bg = list(
            antithetical_modifying_SNP_is_HOM(
                {1: bg}, self.antitheticals
            ).values()
        )[0]
        self.assertTrue(self.pair1 in filtered_bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair1
            not in filtered_bg.filtered_out[
                "antithetical_modifying_SNP_is_HOM"
            ]
        )

    def test_empty_normal_list(self):
        bg = BloodGroup(
            type="LU",
            alleles={AlleleState.NORMAL: []},
            sample="128",
            variant_pool={"19:45315445_G_A": Zygosity.HET},
            filtered_out=defaultdict(list),
        )
        filtered_bg = list(
            antithetical_modifying_SNP_is_HOM(
                {1: bg}, self.antitheticals
            ).values()
        )[0]
        self.assertEqual(filtered_bg.alleles[AlleleState.NORMAL], [])



class TestCantPairWithRefCuzSNPsMustBeOnOtherSide(unittest.TestCase):
    def setUp(self):
        self.allele1 = Allele(
            genotype="JK*01",
            phenotype="WIP",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"18:43319519_ref"}),
            null=False,
            weight_geno=1000,
            reference=True,
            sub_type="JK*01",
        )
        self.allele2 = Allele(
            genotype="JK*01W.03",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"18:43310313_G_A"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="JK*01",
        )
        self.allele3 = Allele(
            genotype="JK*01W.04",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"18:43311054_G_A"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="JK*01",
        )
        self.allele4 = Allele(
            genotype="JK*01W.11",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"18:43310313_G_A", "18:43311054_G_A"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="JK*01",
        )

        self.pair1 = Pair(
            allele1=self.allele3, allele2=self.allele2
        )  # JK*01W.03/4 can be on oposite strands - ok
        self.pair2 = Pair(
            allele1=self.allele1, allele2=self.allele2
        )  # JK*01W.03/4 can't be paired with ref as that means 18:43310313_G_A and
        # 18:43311054_G_A are together, which equals JK*01W.11 - not ok
        self.pair3 = Pair(
            allele1=self.allele1, allele2=self.allele4
        )  # JK*01W.11 can be with ref - ok
        self.bg = BloodGroup(
            type="JK",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2, self.pair3]},
            sample="003Kenya",
            variant_pool={
                "18:43310313_G_A": "Heterozygous",
                "18:43311054_G_A": "Heterozygous",
                "18:43311131_G_A": "Heterozygous",
                "18:43316538_A_G": "Heterozygous",
            },
            filtered_out=defaultdict(list),
        )
        cant_pair_with_ref_cuz_SNPs_must_be_on_other_side({1: self.bg})

    def test_pairs_removed(self):
        self.assertTrue(self.pair2 not in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair2
            in self.bg.filtered_out["cant_pair_with_ref_cuz_SNPs_must_be_on_other_side"]
        )

    def test_pairs_not_removed(self):
        self.assertTrue(self.pair1 in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair1
            not in self.bg.filtered_out[
                "cant_pair_with_ref_cuz_SNPs_must_be_on_other_side"
            ]
        )
        self.assertTrue(self.pair3 in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair3
            not in self.bg.filtered_out[
                "cant_pair_with_ref_cuz_SNPs_must_be_on_other_side"
            ]
        )


class TestABOCantPairWithRefCuz261delGHET(unittest.TestCase):
    def setUp(self):
        self.allele1 = Allele(
            genotype="ABO*A1.01",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"9:136132908_T_TC"}),
            null=False,
            weight_geno=1000,
            reference=True,
            sub_type="ABO*A",
        )
        self.allele2 = Allele(
            genotype="ABO*AW.25",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset(
                {
                    "9:136131056_CG_C",
                    "9:136131289_C_T",
                    "9:136131651_G_A",
                    "9:136132908_T_TC",
                }
            ),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="ABO*A",
        )
        self.allele3 = Allele(
            genotype="ABO*O.01.05",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({"9:136132873_T_C", "9:136132908_ref"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="ABO*O",
        )

        self.pair1 = Pair(
            allele1=self.allele1, allele2=self.allele2
        )  # Not possible as 136132908_T_TC is in defining vars so need an O - not ok
        self.pair2 = Pair(allele1=self.allele1, allele2=self.allele3)  # ok

        self.bg = BloodGroup(
            type="ABO",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2]},
            sample="192",
            variant_pool={
                "9:136132908_T_TC": "Heterozygous",
                "9:136132908_ref": "Heterozygous",
                "9:136131056_CG_C": "Heterozygous",
                "9:136131289_C_T": "Heterozygous",
                "9:136131651_G_A": "Heterozygous",
                "9:136132873_T_C": "Heterozygous",
            },
            filtered_out=defaultdict(list),
        )
        ABO_cant_pair_with_ref_cuz_261delG_HET({1: self.bg})

    def test_pairs_removed(self):
        self.assertTrue(self.pair2 in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair2
            not in self.bg.filtered_out["ABO_cant_pair_with_ref_cuz_261delG_HET"]
        )

    def test_pairs_not_removed(self):
        self.assertTrue(self.pair1 not in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair1 in self.bg.filtered_out["ABO_cant_pair_with_ref_cuz_261delG_HET"]
        )


class TestABOCantPairWithRefCuzTrumped(unittest.TestCase):
    def setUp(self):
        self.allele1 = Allele(
            genotype="FUT3*01",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({}),
            null=False,
            weight_geno=1000,
            reference=True,
            sub_type="FUT3*01",
        )
        self.allele2 = Allele(
            genotype="FUT3*01.16",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset(
                {"19:5844043_C_T", "19:5844184_C_T", "19:5844367_C_T"}
            ),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="FUT3*01",
        )
        self.allele3 = Allele(
            genotype="FUT3*01N.01.02",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset(
                {"19:5844184_C_T", "19:5844367_C_T", "19:5844838_C_T"}
            ),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="FUT3*01",
        )

        self.pair1 = Pair(
            allele1=self.allele1, allele2=self.allele2
        )  # Not possible - not ok
        self.pair2 = Pair(allele1=self.allele1, allele2=self.allele3)  # ok

        self.bg = BloodGroup(
            type="FUT3",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2]},
            sample="126",
            variant_pool={
                "19:5843883_C_G": "Heterozygous",
                "19:5844043_C_T": "Heterozygous",
                "19:5844184_C_T": "Heterozygous",
                "19:5844367_C_T": "Homozygous",
                "19:5844838_C_T": "Homozygous",
            },
            filtered_out=defaultdict(list),
        )
        cant_pair_with_ref_cuz_trumped({1: self.bg})

    def test_pairs_not_removed(self):
        self.assertTrue(self.pair2 in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair2 not in self.bg.filtered_out["cant_pair_with_ref_cuz_trumped"]
        )

    def test_pairs_removed(self):
        self.assertTrue(self.pair1 not in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair1 in self.bg.filtered_out["cant_pair_with_ref_cuz_trumped"]
        )


class TestABOCantPairWithRefCuzTrumped2(unittest.TestCase):
    def setUp(self):
        self.allele1 = Allele(
            genotype="FUT3*01",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset({}),
            null=False,
            weight_geno=1000,
            reference=True,
            sub_type="FUT3*01",
        )
        self.allele2 = Allele(
            genotype="FUT3*01.16",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset(
                {"19:5844043_C_T", "19:5844184_C_T", "19:5844367_C_T"}
            ),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="FUT3*01",
        )
        self.allele3 = Allele(
            genotype="FUT3*01N.01.02",
            phenotype=".",
            genotype_alt=".",
            phenotype_alt=".",
            defining_variants=frozenset(
                {"19:5844184_C_T", "19:5844367_C_T", "19:5844838_C_T"}
            ),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="FUT3*01",
        )

        self.pair1 = Pair(allele1=self.allele1, allele2=self.allele2)  # 2x HET - ok
        self.pair2 = Pair(allele1=self.allele1, allele2=self.allele3)  # ok

        self.bg = BloodGroup(
            type="FUT3",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2]},
            sample="126",
            variant_pool={
                "19:5843883_C_G": "Heterozygous",
                "19:5844043_C_T": "Heterozygous",
                "19:5844184_C_T": "Heterozygous",
                "19:5844367_C_T": "Heterozygous",
                "19:5844838_C_T": "Homozygous",
            },
            filtered_out=defaultdict(list),
        )
        cant_pair_with_ref_cuz_trumped({1: self.bg})

    def test_pairs_not_removed(self):
        self.assertTrue(self.pair1 in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair1 not in self.bg.filtered_out["cant_pair_with_ref_cuz_trumped"]
        )
        self.assertTrue(self.pair2 in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair2 not in self.bg.filtered_out["cant_pair_with_ref_cuz_trumped"]
        )




class TestFilterHETPairsByWeight(unittest.TestCase):
    def setUp(self):
        self.allele1 = Allele(
            genotype="FUT2*01",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"19:49206250_ref"}),
            null=False,
            weight_geno=1000,
            reference=True,
            sub_type="FUT2*01",
        )
        self.allele2 = Allele(
            genotype="FUT2*01.03.01",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"19:49206286_A_G"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="FUT2*01",
        )
        self.allele3 = Allele(
            genotype="FUT2*01N.02",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"19:49206674_G_A"}),
            null=False,
            weight_geno=1,
            reference=False,
            sub_type="FUT2*01",
        )
        self.allele4 = Allele(
            genotype="FUT2*01N.16",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"19:49206985_G_A"}),
            null=False,
            weight_geno=8,
            reference=False,
            sub_type="FUT2*01",
        )

        self.pair1 = Pair(allele1=self.allele1, allele2=self.allele2)  # not ok
        self.pair2 = Pair(allele1=self.allele1, allele2=self.allele4)  # not ok
        self.pair3 = Pair(allele1=self.allele1, allele2=self.allele3)  # ok
        self.bg = BloodGroup(
            type="FUT2",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2, self.pair3]},
            sample="001Kenya",
            variant_pool={
                "19:49206250_ref": "Homozygous",
                "19:49206286_A_G": "Heterozygous",
                "19:49206674_G_A": "Heterozygous",
                "19:49206985_G_A": "Heterozygous",
            },
            filtered_out=defaultdict(list),
        )
        filter_HET_pairs_by_weight({1: self.bg})

    def test_pairs_not_removed(self):
        self.assertTrue(self.pair3 in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair3 not in self.bg.filtered_out["filter_HET_pairs_by_weight"]
        )

    def test_pairs_removed(self):
        self.assertTrue(self.pair1 not in self.bg.alleles[AlleleState.NORMAL])
        self.assertTrue(
            self.pair2 in self.bg.filtered_out["filter_HET_pairs_by_weight"]
        )


class TestFilterPairsByContext(unittest.TestCase):
    def setUp(self):
        self.allele1 = Allele(
            genotype="A4GALT*01",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"22:43113793_ref"}),
            null=False,
            weight_geno=1000,
            reference=True,
            sub_type="A4GALT*01",
        )
        self.allele2 = Allele(
            genotype="A4GALT*01.02",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"22:43089849_T_C"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="A4GALT*01",
        )
        self.allele3 = Allele(
            genotype="A4GALT*02",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"22:43113793_C_A"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="A4GALT*02",
        )
        self.allele4 = Allele(
            genotype="A4GALT*02.02",
            genotype_alt=".",
            phenotype=".",
            phenotype_alt=".",
            defining_variants=frozenset({"22:43113793_C_A", "22:43089849_T_C"}),
            null=False,
            weight_geno=1000,
            reference=False,
            sub_type="A4GALT*02",
        )

        self.pair1 = Pair(allele1=self.allele1, allele2=self.allele3)  # not ok
        self.pair2 = Pair(allele1=self.allele1, allele2=self.allele4)  # ok
        self.pair3 = Pair(allele1=self.allele2, allele2=self.allele3)
        # not ok (for different reason [antithetical is het])

        self.bg = BloodGroup(
            type="A4GALT",
            alleles={AlleleState.NORMAL: [self.pair1, self.pair2, self.pair3]},
            sample="Kenya",
            variant_pool={
                "22:43089849_T_C": "Heterozygous",
                "22:43113793_C_A": "Heterozygous",
                "22:43113793_ref": "Heterozygous",
            },
            filtered_out=defaultdict(list),
        )
        filter_pairs_by_context({1: self.bg})




if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import operator
from collections import defaultdict
from functools import partial
from rbceq2.core_logic.alleles import Allele, BloodGroup, Pair
from rbceq2.core_logic.constants import LOW_WEIGHT, AlleleState
from rbceq2.core_logic.utils import (
    Zygosity,
    apply_to_dict_values,
    check_available_variants,
    one_HET_variant,
)
from rbceq2.filters.shared_filter_functionality import (
    flatten_alleles,
    proceed,
    all_hom,
    check_var,
)
from icecream import ic

def split_pair_by_ref(pair: Pair) -> tuple[Allele, Allele]:
    """Split a pair of alleles into reference and non-reference.

    This function assumes that exactly one of the alleles in the pair is marked as a
    reference.

    Args:
        pair (Pair): The pair of alleles to split.

    Returns:
        tuple[Allele, Allele]: A tuple where the first element is the reference allele
                                and the second is the non-reference allele.

    Raises:
        ValueError: If both or neither alleles are marked as reference.
    """

    if pair.allele1.reference and not pair.allele2.reference:
        ref, allele = pair.allele1, pair.allele2
    elif not pair.allele1.reference and pair.allele2.reference:
        ref, allele = pair.allele2, pair.allele1
    else:
        raise ValueError("Both ref")

    return ref, allele


@apply_to_dict_values
def filter_pairs_on_antithetical_zygosity(
    bg: BloodGroup, antitheticals: dict[str, list[str]]
) -> BloodGroup:
    """Process genetic data to identify alleles and genotypes.

    Args:
    ----------
    res (dict[str, list[Allele]]):
        A dictionary mapping genotypes to lists of Allele objects.
    variant_pool_numeric (dict[str, int]):
        A dictionary mapping variants to their counts.

    Returns:
    ----------
    list[str]:
        A list of called genotypes based on processed data.

    Example:
    ----------

    hashed ones not possible because loci that defines the antithetical antigens is HET,
    meaning there will def be 1 FY*01* and 1 FY*02*
    (and in this case FY*02/FY*01 not possible as a modifying SNP exists)
    'FY*01N.01/FY*02'
    #'FY*01N.01/FY*01' not possible
    'FY*02N.01/FY*01'
    #'FY*02/FY*01'
    """

    def find_anti_variant_pair(data_list: list[str]) -> tuple[str] | None:
        """
        Parses a list of comma-separated variant strings to find a matching pair.

        A matching pair consists of two variants that share the same chromosomal
        position but have different alleles (one '_ref' and one alternate).
        These positions act as an antithetical switch between sub types

        Args:
            data_list: A list of strings, where each string can contain one or
                    more comma-separated variants.
                    e.g., ['25408711_ref,25390874_ref', '25408711_G_A']

        Returns:
            A tuple containing the reference variant and the alternate variant
            (e.g., ('25408711_ref', '25408711_G_A')).
            Returns None if no such pair is found.
        """
        variants_by_pos = defaultdict(set)
        for item in data_list:
            variants = item.split(",")
            for variant in variants:
                if not variant:
                    continue
                position = variant.split("_")[0]
                variants_by_pos[position].add(variant)

        for position, unique_variants in variants_by_pos.items():
            if len(unique_variants) == 2:
                ref_variant = None
                alt_variant = None
                for variant in unique_variants:
                    if variant.endswith("_ref"):
                        ref_variant = variant
                    else:
                        alt_variant = variant

                if ref_variant and alt_variant:
                    return (ref_variant, alt_variant)
        return None

    to_remove = []

    if bg.type in antitheticals:
        anti1, anti2 = find_anti_variant_pair(antitheticals[bg.type])
        var_pool_no_chrom = {k.split(":")[1]: v for k, v in bg.variant_pool.items()}
        if var_pool_no_chrom.get(anti1) == Zygosity.HOM:
            return bg
        if var_pool_no_chrom.get(anti2) == Zygosity.HOM:
            return bg
        if (
            var_pool_no_chrom.get(anti1) is None
            and var_pool_no_chrom.get(anti2) is None
        ):
            return bg  # KN has alleles without either antithetical SNV

        # remove pairs of same subtype
        for pair in bg.alleles[AlleleState.NORMAL]:
            if pair.allele1.sub_type == pair.allele2.sub_type:
                to_remove.append(pair)
        if to_remove:
            bg.remove_pairs(to_remove, "filter_pairs_on_antithetical_zygosity")

    return bg


@apply_to_dict_values
def antithetical_modifying_SNP_is_HOM(
    bg: BloodGroup, antitheticals: dict[str, list[str]]
) -> BloodGroup:
    """Process genetic data to identify alleles and genotypes.

    Parameters:
    ----------
    res (dict[str, list[Allele]]):
        A dictionary mapping genotypes to lists of Allele objects.
    variant_pool_numeric (dict[str, int]):
        A dictionary mapping variants to their counts.

    Returns
    -------
    list[str]:
        A list of called genotypes based on processed data.

    Example:
    ----------

    'LU*01.19/LU*02' not possible because modifying SNP (45322744_A_G) is hom
    Allele(genotype='LU*02',
            phenotype='LU:2',
            genotype_alt='LU*B',
            phenotype_alt='Lu(a-b+)',
            sample='',
            defining_variants=frozenset({'19:45315445_ref'}),
            weight_geno=LOW_WEIGHT,
            weight_pheno=2,
            reference=True,
            Sub_type='LU*02'),
    Allele(genotype='LU*02.19',
            phenotype='LU:-18,19',
            genotype_alt='.',
            phenotype_alt='Au(a-b+)',
            sample='',
            defining_variants=frozenset({'19:45315445_ref',
                                        '19:45322744_A_G'}), hom
            weight_geno=LOW_WEIGHT,
            weight_pheno=1,
            reference=False,
            Sub_type='LU*02')]},
    sample='128',
    variant_pool={'19:45315445_G_A': 'Heterozygous',
                '19:45315445_ref': 'Heterozygous',
                '19:45322744_A_G': 'Homozygous'},
    genotypes=['LU*01.19/LU*02', 'LU*01.19/LU*02.19'],
    """
    for allele_state in [AlleleState.NORMAL, AlleleState.CO]:
        if not proceed(bg, allele_state):
            continue
        to_remove = []
        if bg.type in antitheticals:
            modifying_SNP = None
            flattened_alleles = flatten_alleles(bg.alleles[allele_state])
            d = defaultdict(set)
            for allele in flattened_alleles:
                if allele.number_of_defining_variants > 1:
                    for variant in allele.defining_variants:
                        if variant.split(":")[1] not in antitheticals[bg.type]:
                            d[allele.sub_type].add(variant)

            if len(d) > 1:
                assert len(d) == 2
                putative_mod_SNPs = set.union(*d.values())
                if len(putative_mod_SNPs) == 1:
                    modifying_SNP = putative_mod_SNPs.pop()
            if (
                modifying_SNP is not None
                and bg.variant_pool[modifying_SNP] == Zygosity.HOM
            ):
                for pair in bg.alleles[allele_state]:
                    for allele in pair:
                        if allele.number_of_defining_variants == 1:
                            variant = list(allele.defining_variants)[0]
                            if variant.split(":")[1] in antitheticals[bg.type]:
                                to_remove.append(pair)

        if to_remove:
            bg.remove_pairs(
                to_remove, "antithetical_modifying_SNP_is_HOM", allele_state
            )
    return bg


@apply_to_dict_values
def cant_pair_with_ref_cuz_SNPs_must_be_on_other_side(bg: BloodGroup) -> BloodGroup:
    """Filter out allele pairs where reference alleles cannot pair with non-
    reference alleles due to SNP strand requirements.

    This function examines allele pairs from the BloodGroup's NORMAL allele
    state. For each pair that contains a reference allele but not all alleles are
    reference, it splits the pair to analyze the non-reference allele. It then
    determines which SNPs must be on the other strand based on the variant pool's
    zygosity annotations. If a non-reference allele's defining variants are all
    present in the set of SNPs that must be on the other strand (or additional
    criteria are met), the pair is removed.

    Args:
        bg (BloodGroup): A BloodGroup object containing allele states, variant pool,
            and other related data.

    Returns:
        BloodGroup: The updated BloodGroup object after filtering out invalid allele
            pairs.

    Example:
        Given allele pairs such as:
          - JK*01W.03 and JK*01W.04 cannot pair with a reference allele due to SNPs
            like '18:43310313_G_A' and '18:43311054_G_A' requiring specific strand
            orientation.
        Such pairs will be removed from the BloodGroup.

    pair: [Allele(genotype='JK*01W.04',
                  phenotype='.',
                  genotype_alt='.',
                  phenotype_alt='Jk(a+ᵂ)',
                  defining_variants=frozenset({'18:43311054_G_A'}),
                  weight_geno=1000,
                  weight_pheno=3,
                  reference=False,
                  sub_type='JK*01',
                  phases=None,
                  number_of_defining_variants=1),
           Allele(genotype='JK*01', #means HOM for 43319519_ref
                  defining_variants=frozenset({'18:43319519_ref'}),
                  reference=True,

    allele3: Allele(genotype='JK*01W.03',
                    defining_variants=frozenset({'18:43310313_G_A'}),

    bg.variant_pool: {'18:43310313_G_A': 'Heterozygous',
                      '18:43311054_G_A': 'Heterozygous',
                      '18:43311131_G_A': 'Heterozygous',
                      '18:43316538_A_G': 'Heterozygous'}
    SNPS_that_need_to_be_on_other_strand: ['18:43310313_G_A']
    flattened_alleles: {Allele(genotype='JK*01W.03',
                               defining_variants=frozenset({'18:43310313_G_A'}),

                        Allele(genotype='JK*01W.04',
                               defining_variants=frozenset({'18:43311054_G_A'}),

                        Allele(genotype='JK*01N.20',
                               defining_variants=frozenset({'18:43310313_G_A',
                                                            '18:43311054_G_A',
                                                            '18:43311131_G_A',
                                                            '18:43316538_A_G'}),

                        Allele(genotype='JK*01W.11',
                               defining_variants=frozenset({'18:43311054_G_A',
                                                      '18:43310313_G_A'}),
    """
    flattened_alleles = {
        allele
        for allele in flatten_alleles(bg.alleles[AlleleState.NORMAL])
        if not allele.reference
    }

    to_remove = []

    # step1 - for non ref allele find which SNPs have to be on other side
    # (because if they were on same side the allele in question woud be something
    # else)
    for pair in bg.alleles[AlleleState.NORMAL]:
        if pair.contains_reference and not pair.all_reference:
            _, allele = split_pair_by_ref(pair)
            for allele2 in flattened_alleles:
                if allele in allele2:
                    SNPS_that_need_to_be_on_other_strand = [
                        SNP
                        for SNP, zygosity in bg.variant_pool.items()
                        if zygosity == Zygosity.HOM
                    ]

                    het_SNPS = []
                    for SNP in allele2.defining_variants.difference(
                        allele.defining_variants
                    ):
                        if bg.variant_pool[SNP] == Zygosity.HET:
                            het_SNPS.append(SNP)
                    if len(het_SNPS) == 1:
                        SNPS_that_need_to_be_on_other_strand += het_SNPS
                        for SNP in allele.defining_variants:
                            for SNP2 in bg.variant_pool:
                                if SNP == SNP2:
                                    continue
                                if SNP.split("_")[0] in SNP2:
                                    SNPS_that_need_to_be_on_other_strand.append(SNP2)
                        # step2 - if the SNPs that must be on the ref side + and
                        # HOMs define anything then ref not possible
                        for allele3 in flattened_alleles:
                            if all(
                                SNP in SNPS_that_need_to_be_on_other_strand
                                for SNP in allele3.defining_variants
                            ):
                                to_remove.append(pair)
    if to_remove:
        bg.remove_pairs(
            to_remove,
            "cant_pair_with_ref_cuz_SNPs_must_be_on_other_side",
        )

    return bg


@apply_to_dict_values
def ABO_cant_pair_with_ref_cuz_261delG_HET(bg: BloodGroup) -> BloodGroup:
    """Ensure ABO allele pairs obey SNP strand constraints for reference alleles.

    When a reference allele is present (defined by the Lane variant 261del,
    i.e. 9:136132908_T_TC in GRCh37), the paired non-reference allele must be an
    O allele. Some pairings (e.g. ABO*A1.01 with ABO*AEL or ABO*AW variants) are
    not allowed if 136132908_T_TC is present in the defining variants. However,
    pairings with ABO*O alleles are allowed.

    Examples:
        ABO*A1.01/ABO*AEL.02 - Not allowed as 136132908_T_TC is in defining vars.
        ABO*A1.01/ABO*AEL.07 - Not allowed as 136132908_T_TC is in defining vars.
        ABO*A1.01/ABO*AW.25 - Not allowed as 136132908_T_TC is in defining vars.
        ABO*A1.01/ABO*AW.31.01 - Not allowed as 136132908_T_TC is in defining vars.
        ABO*A1.01/ABO*O.01.05 - Allowed.
        ABO*A1.01/ABO*O.01.22 - Allowed.
        ABO*A1.01/ABO*O.01.45 - Allowed.
        ABO*A1.01/ABO*O.01.71 - Allowed.

    Args:
        bg (BloodGroup): A BloodGroup object containing allele pairs, a variant
            pool, and other related genetic data.

    Returns:
        BloodGroup: The updated BloodGroup after filtering out allele pairs where a
            reference allele is improperly paired.
    """

    to_remove = []
    for pair in bg.alleles[AlleleState.NORMAL]:
        if pair.contains_reference and not pair.all_reference:
            ref, allele = split_pair_by_ref(pair)
            tmp_pool2 = bg.variant_pool_numeric
            for variant_on_other_strand in ref.defining_variants:
                if variant_on_other_strand in tmp_pool2:
                    tmp_pool2[variant_on_other_strand] -= 1
            check_vars_other_strand = partial(
                check_available_variants, 0, tmp_pool2, operator.gt
            )
            if all(check_vars_other_strand(allele)):
                # ie, can they exist given other strand
                continue
            to_remove.append(pair)

    if to_remove:
        bg.remove_pairs(
            to_remove,
            "ABO_cant_pair_with_ref_cuz_261delG_HET",
        )

    return bg


@apply_to_dict_values
def cant_pair_with_ref_cuz_trumped(bg: BloodGroup) -> BloodGroup:
    """Filter out allele pairs where a reference allele is trumped by a superior allele.

    This function checks allele pairs in the NORMAL state that contain a reference
    allele. If a non-reference allele in the pair is trumped (i.e. has a higher
    weight_geno compared to another allele with one HET variant in the same subtype),
    the pair is removed.


    Args:
        bg (BloodGroup): A BloodGroup object containing allele pairs, variant pool,
            and related genetic data.

    Returns:
        BloodGroup: The updated BloodGroup after filtering out trumped allele pairs.


    Examples:
    ----------
    summary:
        BG Name: FUT3
        Pairs:
            - FUT3*01N.01.02/FUT3*01N.01.12
            - FUT3*01.16/FUT3*01N.01.12
            - FUT3*01/FUT3*01N.01.12
            - FUT3*01/FUT3*01N.01.02
        Filtered Out:
            - FUT3*01/FUT3*01.16 is removed because FUT3*01.16 is trumped by a better allele,
              e.g. FUT3*01N.01.02 or FUT3*01N.01.12.

    BG Name: FUT3
    Pairs:
    Pair(Genotype: FUT3*01N.01.02/FUT3*01N.01.12 Phenotype: ./.)
    Pair(Genotype: FUT3*01.16/FUT3*01N.01.12 Phenotype: ./.)
    Pair(Genotype: FUT3*01/FUT3*01N.01.12 Phenotype: ./.)
    Pair(Genotype: FUT3*01/FUT3*01N.01.02 Phenotype: ./.)

    Filtered Out: defaultdict(<class 'list'>, {'cant_pair_with_ref_cuz_trumped':
    [Pair(Genotype: FUT3*01/FUT3*01.16 Phenotype: ./.)]})

    FUT3*01.16 not with ref due to FUT3*01N.01.02 and FUT3*01N.01.12

    Current truth: FUT3 126

    FUT3*01.16/FUT3*01N.01.12
    FUT3*01/FUT3*01N.01.02
    FUT3*01/FUT3*01N.01.12
    FUT3*01N.01.02/FUT3*01N.01.12
    [Allele(genotype='FUT3*01.15',
            defining_variants=frozenset({'19:5844184_C_T',
                        '19:5844367_C_T'}), HOM
            weight_geno=1000,
            reference=False,
    Allele(genotype='FUT3*01.16',
            defining_variants=frozenset({'19:5844043_C_T',
                                        '19:5844184_C_T',
                                        '19:5844367_C_T'}), HOM
            weight_geno=1000,
            reference=False,
    Allele(genotype='FUT3*01N.01.02',
            defining_variants=frozenset({'19:5844184_C_T',
                                        '19:5844367_C_T', HOM
                                        '19:5844838_C_T'}), HOM
            weight_geno=1,
            reference=False,
    Allele(genotype='FUT3*01N.01.12',
            defining_variants=frozenset({'19:5843883_C_G',
                                        '19:5844367_C_T', HOM
                                        '19:5844838_C_T'}), HOM
            weight_geno=1,
            reference=False,

    res[goi].variant_pool: {'19:5843883_C_G': 'Heterozygous',
                            '19:5844043_C_T': 'Heterozygous',
                            '19:5844184_C_T': 'Heterozygous',
                            '19:5844367_C_T': 'Homozygous',
                            '19:5844838_C_T': 'Homozygous'}
    """

    to_remove = []
    flattened_alleles = flatten_alleles(bg.alleles[AlleleState.NORMAL])

    if not any(allele.reference for allele in flattened_alleles):
        return bg
    alleles_without_ref = [
        allele for allele in flattened_alleles if not allele.reference
    ]

    alleles_with_1_HET_var = [
        allele
        for allele in alleles_without_ref
        if [bg.variant_pool[variant] for variant in allele.defining_variants].count(
            Zygosity.HET
        )
        == 1
    ]
    if alleles_with_1_HET_var:
        for pair in bg.alleles[AlleleState.NORMAL]:
            if pair.contains_reference and not pair.all_reference:
                ref, allele = split_pair_by_ref(pair)
                for allele2 in alleles_with_1_HET_var:
                    if allele.sub_type != allele2.sub_type:
                        continue
                    if allele.weight_geno > allele2.weight_geno:
                        to_remove.append(pair)
    if to_remove:
        bg.remove_pairs(to_remove, "cant_pair_with_ref_cuz_trumped")

    return bg


@apply_to_dict_values
def cant_not_include_null(bg: BloodGroup) -> BloodGroup:
    """Filter out allele pairs where a theres no null, if there's
    a null with < 2 HET variants
    """
    flattened_alleles = flatten_alleles(bg.alleles[AlleleState.NORMAL])
    null_alleles = [allele for allele in flattened_alleles if allele.null]
    null_alleles_with_one_HET_variant = [
        allele for allele in null_alleles if one_HET_variant(allele, bg.variant_pool)
    ]
    if not null_alleles_with_one_HET_variant:
        return bg
    to_remove = []
    for pair in bg.alleles[AlleleState.NORMAL]:
        if pair.allele1.null or pair.allele2.null:
            continue
        to_remove.append(pair)
    if to_remove:
        bg.remove_pairs(to_remove, "cant_not_include_null")

    return bg


@apply_to_dict_values
def ensure_HET_SNP_used(bg: BloodGroup) -> BloodGroup:
    """
    Ensures that heterozygous variants are utilized in allele pairs if they can form
    existing alleles.

    This function iterates over heterozygous variants in the variant pool of a
    BloodGroup object.
    For each heterozygous variant, it checks whether adding this variant to the defining
    variants of alleles in each pair results in alleles that already exist outside the
    pair. If multiple such matches are found, the pair is considered invalid and is
    removed from the allele pairs.

    Args:
        bg (BloodGroup): The BloodGroup object containing allele pairs and variant pool
        information.

    Returns:
        BloodGroup: The updated BloodGroup object with inconsistent allele pairs
        removed.

    Example:
        Suppose you have a BloodGroup object with allele pairs and a variant pool that
        includes heterozygous variants. This function will remove pairs where the
        heterozygous variants are not properly used in the allele definitions, but
        could form existing alleles when added.


        #Results:
        Genotypes count: 2
        Genotypes:
        FUT3*01N.03.02/FUT3*01N.03.02 #not possible as 19:5843969_G_A isn't
        accounted for
        FUT3*01N.03.02/FUT3*01N.03.09
        Phenotypes (numeric):
        Phenotypes (alphanumeric): Le(a-b-)

        #Data:
        Vars:
        19:5844537_ref : Homozygous
        19:5844649_ref : Homozygous
        19:5843969_G_A : Heterozygous
        Vars_phase:

        Vars_phase_set:

        Raw:
        Allele
        genotype: FUT3*01.04
        defining_variants:
                19:5844537_ref
        weight_geno: 1000
        phenotype: . or Active
        reference: False

        Allele
        genotype: FUT3*01N.03.01
        defining_variants:
                19:5844649_ref
        weight_geno: 1
        phenotype: . or Le(a-),Le(b-)
        reference: False

        Allele
        genotype: FUT3*01N.03.02
        defining_variants:
                19:5844649_ref
                19:5844537_ref
        weight_geno: 1
        phenotype: . or Le(a-),Le(b-)
        reference: False

        Allele
        genotype: FUT3*01N.03.09
        defining_variants:
                19:5844649_ref
                19:5843969_G_A
                19:5844537_ref
        weight_geno: 1
        phenotype: . or Le(a-),Le(b-)
        reference: False
    """
    for allele_state in [AlleleState.NORMAL, AlleleState.CO]:
        if not proceed(bg, allele_state):
            continue
        to_remove = []
        for variant, zygo in bg.variant_pool.items():
            if zygo == Zygosity.HET:
                for pair in bg.alleles[allele_state]:
                    if all_hom(bg.variant_pool, pair.allele1) and all_hom(
                        bg.variant_pool, pair.allele2
                    ):
                        hits = check_var(bg, pair, allele_state, variant)
                        if hits:
                            to_remove.append(pair)
        if to_remove:
            bg.remove_pairs(to_remove, "ensure_HET_SNP_used", allele_state)

    return bg


@apply_to_dict_values
def filter_HET_pairs_by_weight(bg: BloodGroup) -> BloodGroup:
    """This filter forced us to make decisions where there was not always
    a clearly correct answer. I have left the example long to show the evolution of
    thought and why we landed where we did.

    Args:
        bg (BloodGroup): The BloodGroup object containing allele pairs and variant pool
        information.

    Returns:
        BloodGroup: The updated BloodGroup object with inconsistent allele pairs
        removed.

    Example
    ----------
     FUT1/2
    All bar ref here is HET
    So while it's possible that any allele can exist in a pair,
    no pair can exist without most weighted allele, if it just has 1 defining SNP
    (or should it be just has 1 HET defining SNP, cuz if the rest HOM, same, right?)
    # update - only if alleles are same subtype! and if the 1 SNP of the most weighted
    # allele is 'in' one of the pair (meaning that most weighted allele doesn't exist?)
    . Ref has to be included (as all SNPS could be together)
    but FUT2*01/FUT2*01.03.01 isn't possible
    Further, FUT2*01/FUT2*01N.16 isn't possible as FUT2*01.03.01 trumps FUT2*01N.16
    [Allele(genotype='FUT2*01',
        genotype_alt='.',
        defining_variants=frozenset({'19:49206250_ref'}), hom
        weight_geno=LOW_WEIGHT,
        weight_pheno=2,
        reference=True,
        Sub_type='FUT2*01'),
    Allele(genotype='FUT2*01.03.01',
        genotype_alt='.',
        defining_variants=frozenset({'19:49206286_A_G'}),
        weight_geno=LOW_WEIGHT,
        weight_pheno=1,
        reference=False,
        Sub_type='FUT2*01'),
    Allele(genotype='FUT2*01N.02',
        genotype_alt='.',
        defining_variants=frozenset({'19:49206674_G_A'}),
        weight_geno=1,
        weight_pheno=5,
        reference=False,
        Sub_type='FUT2*01'),
    Allele(genotype='FUT2*01N.16',
        genotype_alt='.',
        defining_variants=frozenset({'19:49206985_G_A'}),
        weight_geno=8,
        weight_pheno=5,
        reference=False,
        Sub_type='FUT2*01')]

    res["FUT2"].variant_pool: {'19:49206250_ref': 'Homozygous',
                               '19:49206286_A_G': 'Heterozygous',
                               '19:49206674_G_A': 'Heterozygous',
                               '19:49206985_G_A': 'Heterozygous'}


    issue

    [Allele(genotype='JK*01N.19',
                        genotype_alt='.',
                        defining_variants=frozenset({'18:43319274_G_A'}),
                        weight_geno=7,
                        reference=False,
                        sub_type='JK*01',

    Allele(genotype='JK*02N.17',
                        genotype_alt='.',
                        defining_variants=frozenset({'18:43319274_G_A',
                                                    '18:43319519_G_A'}),
                        weight_geno=13,
                        reference=False,
                        sub_type='JK*02',

    Allele(genotype='JK*01W.06',
                        genotype_alt='.',
                        defining_variants=frozenset({'18:43310415_G_A',
                                                    '18:43316538_A_G'}), hom
                        weight_geno=1000,
                        reference=False,
                        sub_type='JK*01',

    JK*02	43319519_G_A

    pool: {'18:43310415_G_A': 'Heterozygous',
                '18:43316538_A_G': 'Homozygous',
                '18:43319274_G_A': 'Heterozygous',
                '18:43319519_G_A': 'Heterozygous'}

    current code removed JK*01W.06/JK*02N.17. At first glance I thought this was the
    correct behaviour
    because JK*01N.19 trumps JK*01W.06 and only has 1 SNP. but I noticed
    that the 1 defining SNP for JK*01N.19 (43319274_G_A) is part of the definition of
    JK*02N.17, which is further complicated by the fact its a different subtype.
    Based on my understanding I think that JK*01W.06/JK*02N.17 is right, because
    the trumpiness of JK*01N.19 is lost to JK*02N.17.

    Is there a clear rule to cover this problem? One the one hand, nulls are always
    the most weighted allele (they trump other options) but at the same time alleles
    whose defining variants are all in the list of defining variants of another alleles
    are no longer considered. Which of these rules is applied first? In this example
    the 1 defining SNP for JK*01N.19 (43319274_G_A) is part of the definition of
    JK*02N.17, so JK*01N.19 'doesn't exist' anymore. That makes sense to me, but if
    JK*01N.19 caused truncation or frameshift then I would say that this should
    be the most important consideration... unless the addition of the other variant to
    make JK*02N.17 rescues from being null? As I think this through I think I'm flipping
    and would think that any SNP that makes null would have to be to first consideration

    assumig we want to handle 'in' and subtype:
        each pair will have just 1 of each subtype (other filters will remove)
        so if pair has to include most weighted allele of subtype (sigh)
    """

    flattened_alleles = flatten_alleles(bg.alleles[AlleleState.NORMAL])
    if not flattened_alleles:
        return bg
    if all(allele.weight_geno == LOW_WEIGHT for allele in flattened_alleles):
        return bg
    sub_types = {allele.sub_type for allele in flattened_alleles}

    keepers = []

    for sub_type in sub_types:
        max_weight = min(
            [
                allele.weight_geno
                for allele in flattened_alleles
                if allele.sub_type == sub_type
            ]
        )  # lowest weight, base 1

        weights_with_1_SNP = [
            allele.weight_geno
            for allele in flattened_alleles
            if one_HET_variant(allele, bg.variant_pool) and allele.sub_type == sub_type
        ]
        max_weight_with_1_SNP = (
            min(weights_with_1_SNP) if weights_with_1_SNP else LOW_WEIGHT
        )

        if max_weight_with_1_SNP == max_weight:
            alleles_with_max_weight_and_1_SNP = [
                allele
                for allele in flattened_alleles
                if one_HET_variant(allele, bg.variant_pool)
                and allele.sub_type == sub_type
                and allele.weight_geno == max_weight
            ]
            for pair in bg.alleles[AlleleState.NORMAL]:
                if all(
                    allele in pair.allele1 or allele in pair.allele2
                    for allele in alleles_with_max_weight_and_1_SNP
                ) or any(
                    allele.weight_geno == max_weight
                    for allele in pair
                    if allele.sub_type == sub_type
                ):
                    keepers.append(pair)
        else:  # if the allele with max weight does't have just 1 het SNP, need to keep all
            for pair in bg.alleles[AlleleState.NORMAL]:
                keepers.append(pair)
    to_remove = [pair for pair in bg.alleles[AlleleState.NORMAL] if pair not in keepers]

    if to_remove:
        bg.remove_pairs(to_remove, "filter_HET_pairs_by_weight")

    return bg


@apply_to_dict_values
def filter_pairs_by_context(bg: BloodGroup) -> BloodGroup:
    """Filter allele pairs by context.

    This function removes allele pairs from the NORMAL state of the BloodGroup if
    the context indicates that the pair cannot exist. It checks whether the
    combination of remaining variant counts and an allele's defining variants
    match those in other alleles. This ensures that pairs like
    'A4GALT*01/A4GALT*02' are filtered out when a more comprehensive allele is
    implied by the available variants. This logic overlaps with
    cant_pair_with_ref_cuz_SNPs_must_be_on_other_side, but is still needed

    Args:
        bg (BloodGroup): A BloodGroup object containing allele pairs and a numeric
            variant pool.

    Returns:
        BloodGroup: The updated BloodGroup with contextually invalid pairs removed.

    Example:
        Given allele definitions:
            - A4GALT*01.02: defining_variants = {'22:43089849_T_C'}
            - A4GALT*02:   defining_variants = {'22:43113793_C_A'}
            - A4GALT*02.02: defining_variants = {'22:43113793_C_A',
                                                 '22:43089849_T_C'}
            - A4GALT*01:   defining_variants = {'22:43113793_ref'}
        And a variant pool:
            {'22:43089849_T_C': 'Heterozygous',
             '22:43113793_C_A': 'Heterozygous'}
        The valid pairs are:
            'A4GALT*01.02/A4GALT*02' and
            'A4GALT*01/A4GALT*02.02',
        while 'A4GALT*01/A4GALT*02' is not possible.
    """
    # need to catch; if remaing var and allele var combine to define another allele
    # and if remaining var defines or combines with other allele to define another
    # allele then pair is not possible, but check that the pair that rules it out
    # does already exist

    to_remove = []
    for pair in bg.alleles[AlleleState.NORMAL]:
        def_vars = {
            a.defining_variants
            for pair2 in bg.alleles[AlleleState.NORMAL]
            for a in pair2
            if a not in pair
        }
        for allele in pair:
            variant_pool_copy = bg.variant_pool_numeric.copy()
            if allele.reference:
                continue

            for variant in allele.defining_variants:
                variant_pool_copy[variant] -= 1
            left_over_vars = [k for k, v in variant_pool_copy.items() if v > 0]
            if any([len(left_over_vars) == 0, len(def_vars) < 2]):
                continue
            remaining = [
                tuple(sorted(set(left_over_vars))),
                tuple(sorted(set(left_over_vars + list(allele.defining_variants)))),
            ]
            if all(variants in def_vars for variants in remaining):
                to_remove.append(pair)
                break
    if to_remove:
        bg.remove_pairs(to_remove, "filter_pairs_by_context")

    return bg


@apply_to_dict_values
def impossible_alleles(bg: BloodGroup) -> BloodGroup:
    """
    Rule out impossible alleles based on Homozygous variants.

    Args:
        bg (BloodGroup): A BloodGroup object containing allele states and phasing
            information.

    Returns:
        bg (BloodGroup): A BloodGroup object containing allele states and phasing
            information.

    Example:

    Allele(genotype='JK*02W.03',
            defining_variants={'18:43310415_G_A',
                                18:43316538_A_G',
                               '18:43319519_G_A'},
    Allele(genotype='JK*02W.04',
           defining_variants={'18:43310415_G_A',
                              '18:43319519_G_A'},

    '18:43310415_G_A': 'Heterozygous',
    '18:43316538_A_G': 'Homozygous',
    '18:43319519_G_A': 'Heterozygous',

    JK*02W.04 is impossible because '18:43316538_A_G' is Homozygous
    """
    for allele_state in [AlleleState.NORMAL, AlleleState.CO]:
        if not proceed(bg, allele_state):
            continue
        alleles = list(flatten_alleles(bg.alleles[allele_state]))
        homs = {
            variant for variant, zygo in bg.variant_pool.items() if zygo == Zygosity.HOM
        }
        impossible_alleles = []
        for allele in alleles:
            for allele2 in alleles:
                if allele.genotype != allele2.genotype and allele2 in allele:
                    dif = allele.defining_variants.difference(allele2.defining_variants)
                    if all(variant in homs for variant in dif):
                        impossible_alleles.append(allele2)

        to_remove = []
        for pair in bg.alleles[allele_state]:
            if pair.allele1 in impossible_alleles or pair.allele2 in impossible_alleles:
                to_remove.append(pair)
        if to_remove:
            bg.remove_pairs(to_remove, "filter_impossible_alleles", allele_state)
    return bg



import gzip
import io
import os
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any
from icecream import ic
import pandas as pd

os.environ["POLARS_MAX_THREADS"] = "7"  # Must be set before polars import
import polars as pl
from loguru import logger

from rbceq2.core_logic.constants import COMMON_COLS, HOM_REF_DUMMY_QUAL


@dataclass(slots=True, frozen=False)
class VCF:
    """A data class to process VCF files for variant calling and analysis.

    Attributes
        input_vcf (Path | pd.DataFrame):
            The input VCF file path or DataFrame.
        lane_variants (dict[str, Any]):
            Mapping of chromosome to variants specific to lanes.
        unique_variants (set[str]):
            A set of unique variant identifiers.
    """

    input_vcf: Path | pd.DataFrame | None
    lane_variants: dict[str, Any]
    unique_variants: set[str]
    sample: str = field(init=False)
    df: pd.DataFrame = field(init=False)
    loci: set[str] = field(init=False)
    variants: dict[str, str] = field(init=False)
    phase_sets: dict[str, dict[int, tuple[int, int]]] = field(init=False)

    def __post_init__(self):
        """Handle initialization after data class creation."""
        object.__setattr__(self, "df", self.handle_single_or_multi())
        object.__setattr__(self, "sample", self.get_sample())
        self.rename_chrom()
        self.remove_home_ref()
        self.encode_variants()
        self.add_loci()
        object.__setattr__(self, "loci", self.set_loci())
        self.add_lane_variants()
        object.__setattr__(self, "variants", self.get_variants())
        self._create_phase_sets()

    def _create_phase_sets(self) -> None:
        """Parses the VCF DataFrame to find phased variants and stores their
        chromosomal ranges (min and max position) by phase set ID.

        This method populates the `self.phase_sets` attribute with a nested
        dictionary: {chromosome: {phase_set_id: (min_position, max_position)}}.
        """
        # Temporary dict to aggregate all positions for each phase set
        # Format: {chrom: {ps_id: [pos1, pos2, ...]}}
        temp_phase_data = {}

        # Filter for rows that might contain phasing info to reduce work
        df_phased = self.df[self.df["FORMAT"].str.contains("PS", na=False)].copy()

        # Convert POS to integer once for performance
        df_phased["POS"] = pd.to_numeric(df_phased["POS"])

        for _, row in df_phased.iterrows():
            format_keys = row["FORMAT"].split(":")
            try:
                ps_index = format_keys.index("PS")
            except ValueError:
                continue  # 'PS' not in this row's FORMAT string

            sample_values = row["SAMPLE"].split(":")
            genotype = sample_values[0]

            # Ensure genotype is phased ('|') and the PS value is valid
            if "|" in genotype and len(sample_values) > ps_index:
                ps_value = sample_values[ps_index]
                if ps_value != ".":
                    chrom = row["CHROM"]
                    pos = row["POS"]
                    ps_id = int(ps_value)

                    # Initialize nested dicts if they don't exist and append pos
                    temp_phase_data.setdefault(chrom, {}).setdefault(ps_id, []).append(
                        pos
                    )
                   

        # Convert the lists of positions to (min, max) tuples
        final_phase_sets = {}
        for chrom, ps_groups in temp_phase_data.items():
            final_phase_sets[chrom] = {
                ps_id: (min(positions), max(positions))
                for ps_id, positions in ps_groups.items()
            }

        object.__setattr__(self, "phase_sets", final_phase_sets)

    def handle_single_or_multi(self) -> pd.DataFrame:
        """Handle single or multiple entries in the VCF, returning a DataFrame.

        Returns:
            pd.DataFrame: The DataFrame representation of the VCF data.
        """
        if isinstance(self.input_vcf, Path):
            vcf = read_vcf(self.input_vcf)
            return filter_VCF_to_BG_variants(vcf, self.unique_variants)
        else:
            return self.input_vcf[0]

    def rename_chrom(self) -> None:
        """Rename chromosome identifiers by removing the 'chr' prefix."""
        self.df["CHROM"] = self.df["CHROM"].apply(lambda x: x.replace("chr", ""))

    def remove_home_ref(self) -> None:
        """Remove homozygous reference calls from the DataFrame."""
        self.df = self.df[~self.df["SAMPLE"].str.startswith("0/0")].copy(deep=True)

    def encode_variants(self) -> None:
        """Encode variants into a unified format in the DataFrame."""

        def join_vars(chrom: str, pos: str, ref: str, alts: str) -> bool:
            return ",".join([f"{chrom}:{pos}_{ref}_{alt}" for alt in alts.split(",")])

        self.df["variant"] = self.df.apply(
            lambda x: join_vars(
                x["CHROM"],
                x["POS"],
                x["REF"],
                x["ALT"],
            ),
            axis=1,
        )

    def add_loci(self) -> None:
        """Add loci identifiers to the DataFrame."""
        self.df["loci"] = self.df.CHROM + ":" + self.df.POS

    def get_sample(self) -> str:
        """Determine the sample name from the VCF path or DataFrame.

        Returns:
            str: The sample name.
        """
        if isinstance(self.input_vcf, Path):
            return self.input_vcf.stem
        else:
            return self.input_vcf[-1]

    def set_loci(self) -> set[str]:
        """Create a set of loci identifiers from the DataFrame.

        Returns:
            set[str]: The set of loci identifiers.
        """
        return set(self.df.loci)

    def add_lane_variants(self) -> None:
        """Add lane-specific variants to the DataFrame based on existing loci,
        where Lane variants are those of the type first brought to my attention
        by a paper by Dr. Lane. they are those where a variant in the context
        of a given transcript is just wildtype in a genomic reference. ie
        GRCh37/8

        Example:

        1 - new lanes - HOM loci:
        Generic middle cols = ... =
        ID  REF  ALT  QUAL  FILTER  INFO  GT:GQ:DP:AD:AF:PS  ./.:3,89:92:99:0.967:.

        1   25643553  ...   1:25643553_ref  loci

        2 - HET at loci:
        1  159175354   G   A  ... 1:159175354_G_A  1:159175354
        Becomes:
        1  159175354   G   A  ... 1:159175354_G_A,1:159175354_ref  1:159175354

        """

        new_lanes = {}

        for chrom, loci in self.lane_variants.items():
            # assert 'CHR' not in chrom.upper() #del after a few months of use TODO
            chrom = chrom.replace("chr", "")
            for pos in loci:
                # TODO blindly adding is problematic,  what if there's just no read
                # depth - vcf should be forced to report these
                lane_loci = f"{chrom}:{pos}"
                if lane_loci in self.loci:
                    GT = (
                        self.df.loc[self.df.loci == lane_loci, "SAMPLE"]
                        .values[0]
                        .split(":")[0]
                    )
                    assert GT.count("/") == 1 or GT.count("|") == 1
                    assert (
                        "2" not in GT
                    )  # these 2 asserts are designed to find examples so I can sort
                    # complex/multi var
                    if GT.startswith(("0/1", "0|1", "1/0", "1|0")):
                        self.df.loc[self.df.loci == lane_loci, "variant"] = (
                            self.df.loc[self.df.loci == lane_loci, "variant"].values[0]
                            + f",{lane_loci}_ref"
                        )
                else:
                    new_lanes[lane_loci] = (
                        [chrom, pos]
                        + COMMON_COLS[2:-1]
                        + ["GT:AD:GQ:DP:PS"]
                        + [
                            HOM_REF_DUMMY_QUAL,
                            f"{lane_loci}_ref",
                            "loci",
                        ]
                    )
        if new_lanes:
            new_lanes_df = pd.DataFrame.from_dict(new_lanes, orient="index")
            new_lanes_df.columns = self.df.columns
            self.df = pd.concat([self.df, new_lanes_df])

    def get_variants(self) -> dict[str, str]:
        """Retrieve variant information from the DataFrame.

        Returns:
            dict[str, str]: A dictionary of variants and their associated metrics.
        """
        vcf_variants = {}
        for varaint, metrics, format in zip(
            list(self.df.variant), list(self.df["SAMPLE"]), list(self.df["FORMAT"])
        ):
            if isinstance(metrics, float):
                continue
            assert format.startswith("GT")  # needed for add_lane_variants
            mapped_metrics = dict(
                zip(format.strip().split(":"), metrics.strip().split(":"))
            )
            mapped_metrics["GT"] == mapped_metrics["GT"].replace("|", "/")
            if mapped_metrics["GT"] == "0/0":
                continue
            if "," in varaint:
                for variant in varaint.split(","):
                    vcf_variants[variant] = mapped_metrics
            else:
                vcf_variants[varaint] = mapped_metrics

        return vcf_variants


def split_vcf_to_dfs(vcf_df: pd.DataFrame) -> pd.DataFrame:
    """Split multi-sample VCF DataFrame into individual sample DataFrames.

    Args:
        vcf_df (pd.DataFrame): Multi-sample VCF loaded into a DataFrame.

    Returns:
        Dict[str, pd.DataFrame]: Dictionary of sample-specific DataFrames.
    """
    # Extract column names related to samples
    sample_cols = [col for col in vcf_df.columns if col not in COMMON_COLS]

    for sample in sample_cols:
        try:
            assert all(row[1] in ("|", "/") for row in vcf_df[sample])
        except TypeError:
            logger.info(f"Sample {sample} is not diploid")
        cols: list[str] = COMMON_COLS + [sample]
        sample_vcf_df = vcf_df[cols].copy(deep=True)
        sample_vcf_df.columns = COMMON_COLS + ["SAMPLE"]
        yield sample_vcf_df, sample


def find_phased_neighbors(df: pd.DataFrame) -> set[str]:
    """
    rescues ABOs c.261delG - indels don't always get phased properly"""
    central_loci_to_find = {
        "9:133257521": 133257521,
        "9:136132908": 136132908,
    }

    # 1. Create the sorted list of all PHASED loci on Chromosome 9
    phased_on_chrom9 = (
        df.filter(pl.col("CHROM") == "9")
        .with_columns(pl.col("POS").cast(pl.Int64))
        .filter(pl.col("FORMAT").str.contains("PS"))
        .sort("POS")
    )

    # If there are no phased loci at all, exit
    if phased_on_chrom9.height == 0:
        return set()

    # Extract the positions and loci as separate series for quick lookups
    phased_positions = phased_on_chrom9.get_column("POS")
    phased_loci_series = phased_on_chrom9.get_column("LOCI")

    results = []
    # 2. For each central locus, find its place in the sorted list
    for locus_id, locus_pos in central_loci_to_find.items():
        # search_sorted finds the index where `locus_pos` would be inserted
        # to maintain the sort order. This is the index of the first variant
        # at or after our central locus.
        idx = phased_positions.search_sorted(locus_pos)

        # 3. Use this index to find the neighbors from our list of phased loci
        # We need to be careful about edges (e.g., asking for index -2 when idx is 0 or 1)
        results.append(
            {
                "LOCI": locus_id,
                "prev_2": phased_loci_series[idx - 2] if idx > 1 else None,
                "prev_1": phased_loci_series[idx - 1] if idx > 0 else None,
                "next_1": phased_loci_series[idx]
                if idx < len(phased_loci_series)
                else None,
                "next_2": phased_loci_series[idx + 1]
                if idx < len(phased_loci_series) - 1
                else None,
            }
        )
    neighbor_cols = ["prev_2", "prev_1", "next_1", "next_2"]
    neighbours_df = pl.from_dicts(results)
    # 4. Convert the list of dictionaries to a final DataFrame
    unique_loci_set = {
    locus
    for row in neighbours_df.select(neighbor_cols).rows()
    for locus in row
    if locus is not None  # Filter out the nulls
    }
    return unique_loci_set


def filter_VCF_to_BG_variants(df: pl.DataFrame, unique_variants) -> pd.DataFrame:
    """Filter a VCF represented as a Polars DataFrame to only include specified variants.

    This function creates a temporary column 'LOCI' by concatenating the 'CHROM' and
    'POS' columns, filters the DataFrame to retain only rows where 'LOCI' is in the
    provided unique_variants list, converts the result to a Pandas DataFrame, and
    removes the temporary 'LOCI' column.

    Args:
        df (pl.DataFrame): A Polars DataFrame containing VCF file data with columns
            such as "CHROM" and "POS".
        unique_variants (list[str]): A list of unique variant identifiers (e.g.,
            "chr:pos") to filter the DataFrame.

    Returns:
        pd.DataFrame: A Pandas DataFrame containing only the filtered variants from
            the original DataFrame, with the temporary 'LOCI' column removed.
    """
    df = df.with_columns(
        pl.concat_str(pl.col("CHROM"), pl.lit(":"), pl.col("POS")).alias("LOCI")
    )
    neighbours = find_phased_neighbors(df)
    merged_set = neighbours | unique_variants
    filtered_df = df.filter(pl.col("LOCI").is_in(merged_set))
    if filtered_df.height == 0:  # empty
        pandas_df = df.to_pandas(use_pyarrow_extension_array=False)
    else:
        pandas_df = filtered_df.to_pandas(use_pyarrow_extension_array=False)

    del pandas_df["LOCI"]

    return pandas_df


class VcfMissingHeaderError(Exception):
    """
    Custom exception raised when a VCF file's header is missing,
    empty, or critically invalid (e.g., missing ##fileformat or #CHROM line).
    """

    def __init__(
        self, filename=None, message="VCF header is missing or invalid", reason=None
    ):
        """
        Initializes the VcfMissingHeaderError exception.

        Args:
            filename (str, optional): The path or name of the VCF file. Defaults to None.
            message (str, optional): The base error message.
                                     Defaults to "VCF header is missing or invalid".
            reason (str, optional): A specific reason for the header failure
                                    (e.g., "File is empty",
                                     "Missing '##fileformat' line",
                                     "Missing '#CHROM' line"). Defaults to None.
        """
        self.filename = filename
        self.base_message = message
        self.reason = reason

        full_message = self.base_message
        if filename:
            base_filename = os.path.basename(filename)
            full_message += f" in file: '{base_filename}'"
        if reason:
            full_message += f". Reason: {reason}"

        super().__init__(full_message)


class VcfNoDataError(Exception):
    """
    Custom exception raised when a VCF file is found to contain no
    variant data records (potentially only a header).
    """

    def __init__(self, filename=None, message="VCF contains no data records"):
        """
        Initializes the VcfNoDataError exception.

        Args:
            filename (str, optional): The path or name of the VCF file. Defaults to None.
            message (str, optional): The base error message.
                                     Defaults to "VCF contains no data records".
        """
        self.filename = filename
        self.message = message

        if filename:
            base_filename = os.path.basename(filename)
            full_message = f"{message} in file: '{base_filename}'"
        else:
            full_message = message

        super().__init__(full_message)

    def __str__(self):
        return super().__str__()


def read_vcf(file_path: str) -> pl.DataFrame:
    """Read a VCF file using polars while preserving the header and sample names.

    This function manually extracts the header (line starting with "#CHROM")
    and skips meta-information lines (starting with "##"). It then constructs a
    CSV-formatted string and parses it with polars.

    Args:
        file_path (str): Path to the VCF file (can be gzipped).

    Returns:
        pl.DataFrame: DataFrame containing the VCF data.
    """
    header = None
    # Use gzip.open if file is gzipped, else standard open.
    open_func = gzip.open if str(file_path).endswith(".gz") else open
    with open_func(file_path, "rt") as f:
        # Find header line starting with "#CHROM"
        for line in f:
            if line.startswith("##"):
                continue
            if line.startswith("#"):
                header = line.lstrip("#").strip().split("\t")
                if len(header) == 10:
                    header[-1] = "SAMPLE"  # for single sample
                break

        if header is None:
            raise VcfMissingHeaderError(filename=file_path)

        header_line = "\t".join(header) + "\n"
        data = f.read()

    csv_content = header_line + data
    df = pl.read_csv(
        io.StringIO(csv_content),
        separator="\t",
        schema_overrides=dict.fromkeys(["CHROM", "POS", "QUAL"], str),
    )
    if df.is_empty():
        raise VcfNoDataError(filename=file_path)
    df = df.with_columns(df["CHROM"].str.replace("chr", "", literal=True))
    return df


def check_if_multi_sample_vcf(file_path: str) -> bool:
    """Read a VCF file header.

    This function manually extracts the header (line starting with "#CHROM")
    to check if multi sample

    Args:
        file_path (str): Path to the VCF file (can be gzipped).

    Returns:
        bool: True if there's multiple samples

    """
    header = None
    # Use gzip.open if file is gzipped, else standard open.
    open_func = gzip.open if str(file_path).endswith(".gz") else open
    with open_func(file_path, "rt") as f:
        # Find header line starting with "#CHROM"
        for line in f:
            if line.startswith("##"):
                continue
            if line.startswith("#"):
                header = line.lstrip("#").strip().split("\t")
                if len(header) == 10:
                    return False
                elif len(header) < 10:
                    raise VcfMissingHeaderError(filename=file_path)
                else:
                    assert len(header) == len(set(header))
                    # unique sample names
            break

    return True


# backup while adding phase sets
# import gzip
# import io
# import os
# from dataclasses import dataclass, field
# from pathlib import Path
# from typing import Any

# import pandas as pd

# os.environ["POLARS_MAX_THREADS"] = "7"  # Must be set before polars import
# import polars as pl
# from loguru import logger

# from rbceq2.core_logic.constants import COMMON_COLS, HOM_REF_DUMMY_QUAL


# @dataclass(slots=True, frozen=False)
# class VCF:
#     """A data class to process VCF files for variant calling and analysis.

#     Attributes
#         input_vcf (Path | pd.DataFrame):
#             The input VCF file path or DataFrame.
#         lane_variants (dict[str, Any]):
#             Mapping of chromosome to variants specific to lanes.
#         unique_variants (set[str]):
#             A set of unique variant identifiers.
#     """

#     input_vcf: Path | pd.DataFrame | None
#     lane_variants: dict[str, Any]
#     unique_variants: set[str]
#     sample: str = field(init=False)
#     df: pd.DataFrame = field(init=False)
#     loci: set[str] = field(init=False)
#     variants: dict[str, str] = field(init=False)
#     #add phases here

#     def __post_init__(self):
#         """Handle initialization after data class creation."""
#         object.__setattr__(self, "df", self.handle_single_or_multi())
#         object.__setattr__(self, "sample", self.get_sample())
#         self.rename_chrom()
#         self.remove_home_ref()
#         self.encode_variants()
#         self.add_loci()
#         object.__setattr__(self, "loci", self.set_loci())
#         self.add_lane_variants()
#         object.__setattr__(self, "variants", self.get_variants())
#         #add func to add phases here

#     def handle_single_or_multi(self) -> pd.DataFrame:
#         """Handle single or multiple entries in the VCF, returning a DataFrame.

#         Returns:
#             pd.DataFrame: The DataFrame representation of the VCF data.
#         """
#         if isinstance(self.input_vcf, Path):
#             vcf = read_vcf(self.input_vcf)
#             return filter_VCF_to_BG_variants(vcf, self.unique_variants)
#         else:
#             return self.input_vcf[0]

#     def rename_chrom(self) -> None:
#         """Rename chromosome identifiers by removing the 'chr' prefix."""
#         self.df["CHROM"] = self.df["CHROM"].apply(lambda x: x.replace("chr", ""))

#     def remove_home_ref(self) -> None:
#         """Remove homozygous reference calls from the DataFrame."""
#         self.df = self.df[~self.df["SAMPLE"].str.startswith("0/0")].copy(deep=True)

#     def encode_variants(self) -> None:
#         """Encode variants into a unified format in the DataFrame."""

#         def join_vars(chrom: str, pos: str, ref: str, alts: str) -> bool:
#             return ",".join([f"{chrom}:{pos}_{ref}_{alt}" for alt in alts.split(",")])

#         self.df["variant"] = self.df.apply(
#             lambda x: join_vars(
#                 x["CHROM"],
#                 x["POS"],
#                 x["REF"],
#                 x["ALT"],
#             ),
#             axis=1,
#         )

#     def add_loci(self) -> None:
#         """Add loci identifiers to the DataFrame."""
#         self.df["loci"] = self.df.CHROM + ":" + self.df.POS

#     def get_sample(self) -> str:
#         """Determine the sample name from the VCF path or DataFrame.

#         Returns:
#             str: The sample name.
#         """
#         if isinstance(self.input_vcf, Path):
#             return self.input_vcf.stem
#         else:
#             return self.input_vcf[-1]

#     def set_loci(self) -> set[str]:
#         """Create a set of loci identifiers from the DataFrame.

#         Returns:
#             set[str]: The set of loci identifiers.
#         """
#         return set(self.df.loci)

#     def add_lane_variants(self) -> None:
#         """Add lane-specific variants to the DataFrame based on existing loci,
#         where Lane variants are those of the type first brought to my attention
#         by a paper by Dr. Lane. they are those where a variant in the context
#         of a given transcript is just wildtype in a genomic reference. ie
#         GRCh37/8

#         Example:

#         1 - new lanes - HOM loci:
#         Generic middle cols = ... =
#         ID  REF  ALT  QUAL  FILTER  INFO  GT:GQ:DP:AD:AF:PS  ./.:3,89:92:99:0.967:.

#         1   25643553  ...   1:25643553_ref  loci

#         2 - HET at loci:
#         1  159175354   G   A  ... 1:159175354_G_A  1:159175354
#         Becomes:
#         1  159175354   G   A  ... 1:159175354_G_A,1:159175354_ref  1:159175354

#         """

#         new_lanes = {}

#         for chrom, loci in self.lane_variants.items():
#             # assert 'CHR' not in chrom.upper() #del after a few months of use TODO
#             chrom = chrom.replace("chr", "")
#             for pos in loci:
#                 # TODO blindly adding is problematic,  what if there's just no read
#                 # depth - vcf should be forced to report these
#                 lane_loci = f"{chrom}:{pos}"
#                 if lane_loci in self.loci:
#                     GT = (
#                         self.df.loc[self.df.loci == lane_loci, "SAMPLE"]
#                         .values[0]
#                         .split(":")[0]
#                     )
#                     assert GT.count("/") == 1 or GT.count("|") == 1
#                     assert (
#                         "2" not in GT
#                     )  # these 2 asserts are designed to find examples so I can sort
#                     # complex/multi var
#                     if GT.startswith(("0/1", "0|1", "1/0", "1|0")):
#                         self.df.loc[self.df.loci == lane_loci, "variant"] = (
#                             self.df.loc[self.df.loci == lane_loci, "variant"].values[0]
#                             + f",{lane_loci}_ref"
#                         )
#                 else:
#                     new_lanes[lane_loci] = (
#                         [chrom, pos]
#                         + COMMON_COLS[2:-1]
#                         + ["GT:AD:GQ:DP:PS"]
#                         + [
#                             HOM_REF_DUMMY_QUAL,
#                             f"{lane_loci}_ref",
#                             "loci",
#                         ]
#                     )
#         if new_lanes:
#             new_lanes_df = pd.DataFrame.from_dict(new_lanes, orient="index")
#             new_lanes_df.columns = self.df.columns
#             self.df = pd.concat([self.df, new_lanes_df])

#     def get_variants(self) -> dict[str, str]:
#         """Retrieve variant information from the DataFrame.

#         Returns:
#             dict[str, str]: A dictionary of variants and their associated metrics.
#         """
#         vcf_variants = {}
#         for varaint, metrics, format in zip(
#             list(self.df.variant), list(self.df["SAMPLE"]), list(self.df["FORMAT"])
#         ):
#             if isinstance(metrics, float):
#                 continue
#             assert format.startswith("GT")  # needed for add_lane_variants
#             mapped_metrics = dict(
#                 zip(format.strip().split(":"), metrics.strip().split(":"))
#             )
#             mapped_metrics["GT"] == mapped_metrics["GT"].replace("|", "/")
#             if mapped_metrics["GT"] == "0/0":
#                 continue
#             if "," in varaint:
#                 for variant in varaint.split(","):
#                     vcf_variants[variant] = mapped_metrics
#             else:
#                 vcf_variants[varaint] = mapped_metrics

#         return vcf_variants


# def split_vcf_to_dfs(vcf_df: pd.DataFrame) -> pd.DataFrame:
#     """Split multi-sample VCF DataFrame into individual sample DataFrames.

#     Args:
#         vcf_df (pd.DataFrame): Multi-sample VCF loaded into a DataFrame.

#     Returns:
#         Dict[str, pd.DataFrame]: Dictionary of sample-specific DataFrames.
#     """
#     # Extract column names related to samples
#     sample_cols = [col for col in vcf_df.columns if col not in COMMON_COLS]

#     for sample in sample_cols:
#         try:
#             assert all(row[1] in ("|", "/") for row in vcf_df[sample])
#         except TypeError:
#             logger.info(f"Sample {sample} is not diploid")
#         cols: list[str] = COMMON_COLS + [sample]
#         sample_vcf_df = vcf_df[cols].copy(deep=True)
#         sample_vcf_df.columns = COMMON_COLS + ["SAMPLE"]
#         yield sample_vcf_df, sample


# def filter_VCF_to_BG_variants(df: pl.DataFrame, unique_variants) -> pd.DataFrame:
#     """Filter a VCF represented as a Polars DataFrame to only include specified variants.

#     This function creates a temporary column 'LOCI' by concatenating the 'CHROM' and
#     'POS' columns, filters the DataFrame to retain only rows where 'LOCI' is in the
#     provided unique_variants list, converts the result to a Pandas DataFrame, and
#     removes the temporary 'LOCI' column.

#     Args:
#         df (pl.DataFrame): A Polars DataFrame containing VCF file data with columns
#             such as "CHROM" and "POS".
#         unique_variants (list[str]): A list of unique variant identifiers (e.g.,
#             "chr:pos") to filter the DataFrame.

#     Returns:
#         pd.DataFrame: A Pandas DataFrame containing only the filtered variants from
#             the original DataFrame, with the temporary 'LOCI' column removed.
#     """
#     df = df.with_columns(
#         pl.concat_str(pl.col("CHROM"), pl.lit(":"), pl.col("POS")).alias("LOCI")
#     )
#     filtered_df = df.filter(pl.col("LOCI").is_in(unique_variants))
#     if filtered_df.height == 0:  # empty
#         pandas_df = df.to_pandas(use_pyarrow_extension_array=False)
#     else:
#         pandas_df = filtered_df.to_pandas(use_pyarrow_extension_array=False)

#     del pandas_df["LOCI"]

#     return pandas_df


# class VcfMissingHeaderError(Exception):
#     """
#     Custom exception raised when a VCF file's header is missing,
#     empty, or critically invalid (e.g., missing ##fileformat or #CHROM line).
#     """

#     def __init__(
#         self, filename=None, message="VCF header is missing or invalid", reason=None
#     ):
#         """
#         Initializes the VcfMissingHeaderError exception.

#         Args:
#             filename (str, optional): The path or name of the VCF file. Defaults to None.
#             message (str, optional): The base error message.
#                                      Defaults to "VCF header is missing or invalid".
#             reason (str, optional): A specific reason for the header failure
#                                     (e.g., "File is empty",
#                                      "Missing '##fileformat' line",
#                                      "Missing '#CHROM' line"). Defaults to None.
#         """
#         self.filename = filename
#         self.base_message = message
#         self.reason = reason

#         full_message = self.base_message
#         if filename:
#             base_filename = os.path.basename(filename)
#             full_message += f" in file: '{base_filename}'"
#         if reason:
#             full_message += f". Reason: {reason}"

#         super().__init__(full_message)


# class VcfNoDataError(Exception):
#     """
#     Custom exception raised when a VCF file is found to contain no
#     variant data records (potentially only a header).
#     """

#     def __init__(self, filename=None, message="VCF contains no data records"):
#         """
#         Initializes the VcfNoDataError exception.

#         Args:
#             filename (str, optional): The path or name of the VCF file. Defaults to None.
#             message (str, optional): The base error message.
#                                      Defaults to "VCF contains no data records".
#         """
#         self.filename = filename
#         self.message = message

#         if filename:
#             base_filename = os.path.basename(filename)
#             full_message = f"{message} in file: '{base_filename}'"
#         else:
#             full_message = message

#         super().__init__(full_message)

#     def __str__(self):
#         return super().__str__()


# def read_vcf(file_path: str) -> pl.DataFrame:
#     """Read a VCF file using polars while preserving the header and sample names.

#     This function manually extracts the header (line starting with "#CHROM")
#     and skips meta-information lines (starting with "##"). It then constructs a
#     CSV-formatted string and parses it with polars.

#     Args:
#         file_path (str): Path to the VCF file (can be gzipped).

#     Returns:
#         pl.DataFrame: DataFrame containing the VCF data.
#     """
#     header = None
#     # Use gzip.open if file is gzipped, else standard open.
#     open_func = gzip.open if str(file_path).endswith(".gz") else open
#     with open_func(file_path, "rt") as f:
#         # Find header line starting with "#CHROM"
#         for line in f:
#             if line.startswith("##"):
#                 continue
#             if line.startswith("#"):
#                 header = line.lstrip("#").strip().split("\t")
#                 if len(header) == 10:
#                     header[-1] = "SAMPLE"  # for single sample
#                 break

#         if header is None:
#             raise VcfMissingHeaderError(filename=file_path)

#         header_line = "\t".join(header) + "\n"
#         data = f.read()

#     csv_content = header_line + data
#     df = pl.read_csv(
#         io.StringIO(csv_content),
#         separator="\t",
#         schema_overrides=dict.fromkeys(["CHROM", "POS", "QUAL"], str),
#     )
#     if df.is_empty():
#         raise VcfNoDataError(filename=file_path)
#     df = df.with_columns(df["CHROM"].str.replace("chr", "", literal=True))
#     return df


# def check_if_multi_sample_vcf(file_path: str) -> bool:
#     """Read a VCF file header.

#     This function manually extracts the header (line starting with "#CHROM")
#     to check if multi sample

#     Args:
#         file_path (str): Path to the VCF file (can be gzipped).

#     Returns:
#         bool: True if there's multiple samples

#     """
#     header = None
#     # Use gzip.open if file is gzipped, else standard open.
#     open_func = gzip.open if str(file_path).endswith(".gz") else open
#     with open_func(file_path, "rt") as f:
#         # Find header line starting with "#CHROM"
#         for line in f:
#             if line.startswith("##"):
#                 continue
#             if line.startswith("#"):
#                 header = line.lstrip("#").strip().split("\t")
#                 if len(header) == 10:
#                     return False
#                 elif len(header) < 10:
#                     raise VcfMissingHeaderError(filename=file_path)
#                 else:
#                     assert len(header) == len(set(header))
#                     # unique sample names
#             break

#     return True
